"""Unit tests for SchemaValidator class."""

import json
import os
import pytest
import tempfile
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.schema_validator import (
    SchemaValidator,
)
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.validation_utils import (
    ValidationError,
    ValidationResult,
)
from hypothesis import given
from hypothesis import strategies as st


@pytest.mark.unit
class TestSchemaValidator:
    """Unit tests for SchemaValidator class - fast, isolated tests."""

    @pytest.fixture
    def validator(self):
        """Create a SchemaValidator instance for testing."""
        return SchemaValidator()

    def _validate_schema_dict(self, validator, schema_dict):
        """Helper method to validate a schema dictionary."""
        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema_dict, f)
            temp_file = f.name
        try:
            return validator.validate_schema_file(temp_file)
        finally:
            os.unlink(temp_file)

    @pytest.fixture
    def valid_minimal_schema(self):
        """Create a valid minimal schema for testing."""
        return {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                        'sort_key': 'sk',
                    },
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'sk_template': 'ENTITY',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

    def test_validate_valid_schema(self, validator, valid_minimal_schema):
        """Test that a valid minimal schema passes validation."""
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert result.is_valid and len(result.errors) == 0

    def test_validate_schema_not_dict(self, validator):
        """Test that non-dictionary schema fails validation."""
        result = self._validate_schema_dict(validator, [])
        assert not result.is_valid and any(
            'Schema must be a JSON object' in e.message for e in result.errors
        )

    def test_validate_tables_not_list(self, validator):
        """Test that tables must be a list."""
        result = self._validate_schema_dict(validator, {'tables': 'not a list'})
        assert not result.is_valid and any(
            'tables must be an array' in e.message for e in result.errors
        )

    def test_validate_empty_tables(self, validator):
        """Test that tables list cannot be empty."""
        result = self._validate_schema_dict(validator, {'tables': []})
        assert not result.is_valid and any(
            'tables cannot be empty' in e.message for e in result.errors
        )

    def test_validate_table_not_dict(self, validator):
        """Test that each table must be a dictionary."""
        result = self._validate_schema_dict(validator, {'tables': ['not a dict']})
        assert not result.is_valid and any(
            'Table must be an object' in e.message for e in result.errors
        )

    def test_validate_table_config_not_dict(self, validator):
        """Test that table_config must be a dictionary."""
        result = self._validate_schema_dict(
            validator, {'tables': [{'table_config': 'not a dict', 'entities': {}}]}
        )
        assert not result.is_valid and any(
            'table_config must be an object' in e.message for e in result.errors
        )

    def test_validate_entities_not_dict(self, validator):
        """Test that entities must be a dictionary."""
        result = self._validate_schema_dict(
            validator,
            {
                'tables': [
                    {
                        'table_config': {'table_name': 'T', 'partition_key': 'pk'},
                        'entities': 'not a dict',
                    }
                ]
            },
        )
        assert not result.is_valid and any(
            'entities must be an object' in e.message for e in result.errors
        )

    def test_validate_empty_entities(self, validator):
        """Test that entities dictionary cannot be empty."""
        result = self._validate_schema_dict(
            validator,
            {
                'tables': [
                    {'table_config': {'table_name': 'T', 'partition_key': 'pk'}, 'entities': {}}
                ]
            },
        )
        assert not result.is_valid and any(
            'entities cannot be empty' in e.message for e in result.errors
        )

    def test_validate_entity_not_dict(self, validator):
        """Test that each entity must be a dictionary."""
        result = self._validate_schema_dict(
            validator,
            {
                'tables': [
                    {
                        'table_config': {'table_name': 'T', 'partition_key': 'pk'},
                        'entities': {'E': 'not a dict'},
                    }
                ]
            },
        )
        assert not result.is_valid and any('must be an object' in e.message for e in result.errors)

    def test_validate_fields_not_list(self, validator, valid_minimal_schema):
        """Test that fields must be a list."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['fields'] = 'not a list'
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'fields must be an array' in e.message for e in result.errors
        )

    def test_validate_empty_fields(self, validator, valid_minimal_schema):
        """Test that fields list cannot be empty."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['fields'] = []
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'fields cannot be empty' in e.message for e in result.errors
        )

    def test_validate_field_not_dict(self, validator, valid_minimal_schema):
        """Test that each field must be a dictionary."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['fields'] = ['not a dict']
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Field must be an object' in e.message for e in result.errors
        )

    def test_validate_duplicate_field_names(self, validator, valid_minimal_schema):
        """Test that field names must be unique within an entity."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['fields'] = [
            {'name': 'id', 'type': 'string', 'required': True},
            {'name': 'id', 'type': 'string', 'required': False},
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Duplicate field name' in e.message for e in result.errors
        )

    def test_validate_invalid_field_type(self, validator, valid_minimal_schema):
        """Test that field type must be valid."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['fields'][0]['type'] = (
            'invalid_type'
        )
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            "Invalid type value 'invalid_type'" in e.message for e in result.errors
        )

    def test_validate_array_field_missing_item_type(self, validator, valid_minimal_schema):
        """Test that array fields must have item_type specified."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['fields'].append(
            {'name': 'tags', 'type': 'array', 'required': True}
        )
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Array fields must specify item_type' in e.message for e in result.errors
        )

    def test_validate_field_required_not_bool(self, validator, valid_minimal_schema):
        """Test that field required must be boolean."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['fields'][0]['required'] = (
            'yes'
        )
        assert not self._validate_schema_dict(validator, valid_minimal_schema).is_valid

    def test_validate_sk_template_null(self, validator, valid_minimal_schema):
        """Test that sk_template can be null for partition-key-only tables."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['sk_template'] = None
        assert self._validate_schema_dict(validator, valid_minimal_schema).is_valid

    def test_validate_sk_template_invalid_type(self, validator, valid_minimal_schema):
        """Test that sk_template must be string or null."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['sk_template'] = 123
        assert not self._validate_schema_dict(validator, valid_minimal_schema).is_valid

    def test_validate_access_patterns_not_list(self, validator, valid_minimal_schema):
        """Test that access_patterns must be a list."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = (
            'not a list'
        )
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'access_patterns must be an array' in e.message for e in result.errors
        )

    def test_validate_access_pattern_not_dict(self, validator, valid_minimal_schema):
        """Test that each access pattern must be a dictionary."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            'not a dict'
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Access pattern must be an object' in e.message for e in result.errors
        )

    def test_validate_pattern_id_not_int(self, validator, valid_minimal_schema):
        """Test that pattern_id must be an integer."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 'not_an_int',
                'name': 'test',
                'description': 'test',
                'operation': 'GetItem',
                'parameters': [],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'pattern_id must be an integer' in e.message for e in result.errors
        )

    def test_validate_duplicate_pattern_ids(self, validator, valid_minimal_schema):
        """Test that pattern IDs must be unique across all entities."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'p1',
                'description': 't1',
                'operation': 'GetItem',
                'parameters': [],
                'return_type': 'single_entity',
            },
            {
                'pattern_id': 1,
                'name': 'p2',
                'description': 't2',
                'operation': 'GetItem',
                'parameters': [],
                'return_type': 'single_entity',
            },
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Duplicate pattern_id' in e.message for e in result.errors
        )

    def test_validate_duplicate_pattern_names(self, validator, valid_minimal_schema):
        """Test that pattern names must be unique within an entity."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'same',
                'description': 't1',
                'operation': 'GetItem',
                'parameters': [],
                'return_type': 'single_entity',
            },
            {
                'pattern_id': 2,
                'name': 'same',
                'description': 't2',
                'operation': 'GetItem',
                'parameters': [],
                'return_type': 'single_entity',
            },
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Duplicate pattern name' in e.message for e in result.errors
        )

    def test_validate_invalid_enums(self, validator, valid_minimal_schema):
        """Test invalid operation and return_type in one test."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'test',
                'description': 'test',
                'operation': 'InvalidOp',
                'parameters': [],
                'return_type': 'invalid_type',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        assert any('Invalid operation' in e.message for e in result.errors)
        assert any('Invalid return_type' in e.message for e in result.errors)

    def test_validate_parameters_not_list(self, validator, valid_minimal_schema):
        """Test that parameters must be a list."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'test',
                'description': 'test',
                'operation': 'GetItem',
                'parameters': 'not a list',
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'parameters must be an array' in e.message for e in result.errors
        )

    def test_validate_parameter_not_dict(self, validator, valid_minimal_schema):
        """Test that each parameter must be a dictionary."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'test',
                'description': 'test',
                'operation': 'GetItem',
                'parameters': ['not a dict'],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Parameter must be an object' in e.message for e in result.errors
        )

    def test_validate_duplicate_parameter_names(self, validator, valid_minimal_schema):
        """Test that parameter names must be unique within a pattern."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'test',
                'description': 'test',
                'operation': 'GetItem',
                'parameters': [{'name': 'id', 'type': 'string'}, {'name': 'id', 'type': 'string'}],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Duplicate parameter name' in e.message for e in result.errors
        )

    def test_validate_invalid_parameter_type(self, validator, valid_minimal_schema):
        """Test that parameter type must be valid."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'test',
                'description': 'test',
                'operation': 'GetItem',
                'parameters': [{'name': 'id', 'type': 'invalid_type'}],
                'return_type': 'single_entity',
            }
        ]
        assert not self._validate_schema_dict(validator, valid_minimal_schema).is_valid

    def test_validate_entity_parameter_missing_entity_type(self, validator, valid_minimal_schema):
        """Test that entity parameters must have entity_type."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'test',
                'description': 'test',
                'operation': 'PutItem',
                'parameters': [{'name': 'entity', 'type': 'entity'}],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid and any(
            'Entity parameters must specify entity_type' in e.message for e in result.errors
        )

    def test_validate_entity_reference(self, validator):
        """Test that entity references are validated correctly."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'T', 'partition_key': 'pk', 'sort_key': 'sk'},
                    'entities': {
                        'E1': {
                            'entity_type': 'E1',
                            'pk_template': '{id}',
                            'sk_template': 'E1',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [
                                {
                                    'pattern_id': 1,
                                    'name': 'create',
                                    'description': 'test',
                                    'operation': 'PutItem',
                                    'parameters': [
                                        {
                                            'name': 'entity',
                                            'type': 'entity',
                                            'entity_type': 'NonExistent',
                                        }
                                    ],
                                    'return_type': 'single_entity',
                                }
                            ],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        assert not result.is_valid and any(
            "Unknown entity type 'NonExistent'" in e.message for e in result.errors
        )

    def test_validate_duplicate_entity_names_across_tables(self, validator):
        """Test that entity names must be unique across all tables."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'T1', 'partition_key': 'pk'},
                    'entities': {
                        'User': {
                            'entity_type': 'U',
                            'pk_template': '{id}',
                            'sk_template': 'U',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                },
                {
                    'table_config': {'table_name': 'T2', 'partition_key': 'pk'},
                    'entities': {
                        'User': {
                            'entity_type': 'U2',
                            'pk_template': '{id}',
                            'sk_template': 'U',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                },
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        assert not result.is_valid and any(
            "Duplicate entity name 'User' across tables" in e.message for e in result.errors
        )

    def test_validate_file_not_found(self, validator):
        """Test that validation fails gracefully for non-existent files."""
        result = validator.validate_schema_file('/nonexistent/file.json')
        assert not result.is_valid and any(
            'Schema file not found' in e.message for e in result.errors
        )

    def test_validate_invalid_json(self, validator):
        """Test that validation fails gracefully for invalid JSON."""
        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            f.write('{invalid json}')
            temp_file = f.name
        try:
            result = validator.validate_schema_file(temp_file)
            assert not result.is_valid and any('Invalid JSON' in e.message for e in result.errors)
        finally:
            os.unlink(temp_file)

    def test_format_validation_result_success(self, validator):
        """Test formatting of successful validation result."""
        validator.result = ValidationResult(is_valid=True, errors=[], warnings=[])
        assert '✅ Schema validation passed!' in validator.format_validation_result()

    def test_format_validation_result_with_errors_and_warnings(self, validator):
        """Test formatting of validation result with errors and warnings."""
        errors = [
            ValidationError('test.field', 'Test error', 'suggestion'),
            ValidationError('test.other', 'Another error', None),
        ]
        warnings = [ValidationError('test.warning', 'Test warning', None)]
        validator.result = ValidationResult(is_valid=False, errors=errors, warnings=warnings)
        formatted = validator.format_validation_result()
        assert all(
            x in formatted
            for x in [
                '❌ Schema validation failed',
                'Test error',
                'Another error',
                '💡 suggestion',
                '⚠️  Warnings:',
                'Test warning',
            ]
        )

    def test_convenience_function(self):
        """Test the convenience validate_schema_file function."""
        from awslabs.dynamodb_mcp_server.repo_generation_tool.core.schema_validator import (
            validate_schema_file,
        )

        assert not validate_schema_file('/nonexistent/file.json').is_valid

    def test_validate_consistent_read_boolean_type(self, validator, valid_minimal_schema):
        """Test that consistent_read must be a boolean type."""
        # Test with string value
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'get_item',
                'description': 'Get item',
                'operation': 'GetItem',
                'consistent_read': 'true',  # String instead of boolean
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        assert any(
            'consistent_read must be a boolean' in e.message and 'Pattern 1' in e.message
            for e in result.errors
        )

    def test_validate_consistent_read_gsi_restriction(self, validator, valid_minimal_schema):
        """Test that consistent_read cannot be true for GSI queries."""
        # Add GSI to table config
        valid_minimal_schema['tables'][0]['gsi_list'] = [
            {
                'name': 'TestIndex',
                'partition_key': 'gsi_pk',
                'sort_key': 'gsi_sk',
            }
        ]
        # Add GSI mapping to entity
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['gsi_mappings'] = [
            {
                'name': 'TestIndex',
                'pk_template': '{id}',
                'sk_template': 'GSI',
            }
        ]
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'query_by_gsi',
                'description': 'Query by GSI',
                'operation': 'Query',
                'index_name': 'TestIndex',
                'consistent_read': True,  # Not allowed for GSI
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'entity_list',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        assert any(
            'consistent_read cannot be true for GSI queries' in e.message
            and 'Global Secondary Indexes only support eventually consistent reads' in e.message
            for e in result.errors
        )

    def test_validate_consistent_read_gsi_false_allowed(self, validator, valid_minimal_schema):
        """Test that consistent_read: false is allowed for GSI queries."""
        # Add GSI to table config
        valid_minimal_schema['tables'][0]['gsi_list'] = [
            {
                'name': 'TestIndex',
                'partition_key': 'gsi_pk',
                'sort_key': 'gsi_sk',
            }
        ]
        # Add GSI mapping to entity
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['gsi_mappings'] = [
            {
                'name': 'TestIndex',
                'pk_template': '{id}',
                'sk_template': 'GSI',
            }
        ]
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'query_by_gsi',
                'description': 'Query by GSI',
                'operation': 'Query',
                'index_name': 'TestIndex',
                'consistent_read': False,  # Allowed for GSI
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'entity_list',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert result.is_valid

    def test_validate_consistent_read_main_table_true(self, validator, valid_minimal_schema):
        """Test that consistent_read: true is allowed for main table queries."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'get_item',
                'description': 'Get item',
                'operation': 'GetItem',
                'consistent_read': True,  # Allowed for main table
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert result.is_valid

    def test_validate_consistent_read_error_includes_pattern_info(
        self, validator, valid_minimal_schema
    ):
        """Test that error messages include pattern_id and pattern_name."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 42,
                'name': 'my_pattern',
                'description': 'Test pattern',
                'operation': 'GetItem',
                'consistent_read': 123,  # Invalid type
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        assert any('Pattern 42' in e.message and 'my_pattern' in e.message for e in result.errors)

    def test_consistent_read_type_error_includes_pattern_id(self, validator, valid_minimal_schema):
        """Test that type validation error messages include pattern_id."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 99,
                'name': 'test_pattern',
                'description': 'Test pattern',
                'operation': 'GetItem',
                'consistent_read': 'invalid',  # Invalid type
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        # Check that error message includes pattern_id
        assert any('Pattern 99' in e.message for e in result.errors), (
            f"Expected 'Pattern 99' in error messages, got: {[e.message for e in result.errors]}"
        )

    def test_consistent_read_type_error_includes_pattern_name(
        self, validator, valid_minimal_schema
    ):
        """Test that type validation error messages include pattern_name."""
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'get_user_by_id',
                'description': 'Test pattern',
                'operation': 'GetItem',
                'consistent_read': 42,  # Invalid type
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'single_entity',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        # Check that error message includes pattern_name
        assert any('get_user_by_id' in e.message for e in result.errors), (
            f"Expected 'get_user_by_id' in error messages, got: {[e.message for e in result.errors]}"
        )

    def test_consistent_read_gsi_error_includes_pattern_id(self, validator, valid_minimal_schema):
        """Test that GSI restriction error messages include pattern_id."""
        # Add GSI to table config
        valid_minimal_schema['tables'][0]['gsi_list'] = [
            {
                'name': 'TestIndex',
                'partition_key': 'gsi_pk',
                'sort_key': 'gsi_sk',
            }
        ]
        # Add GSI mapping to entity
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['gsi_mappings'] = [
            {
                'name': 'TestIndex',
                'pk_template': '{id}',
                'sk_template': 'GSI',
            }
        ]
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 77,
                'name': 'query_by_gsi',
                'description': 'Query by GSI',
                'operation': 'Query',
                'index_name': 'TestIndex',
                'consistent_read': True,  # Not allowed for GSI
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'entity_list',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        # Check that error message includes pattern_id
        assert any('Pattern 77' in e.message for e in result.errors), (
            f"Expected 'Pattern 77' in error messages, got: {[e.message for e in result.errors]}"
        )

    def test_consistent_read_gsi_error_includes_pattern_name(
        self, validator, valid_minimal_schema
    ):
        """Test that GSI restriction error messages include pattern_name."""
        # Add GSI to table config
        valid_minimal_schema['tables'][0]['gsi_list'] = [
            {
                'name': 'TestIndex',
                'partition_key': 'gsi_pk',
                'sort_key': 'gsi_sk',
            }
        ]
        # Add GSI mapping to entity
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['gsi_mappings'] = [
            {
                'name': 'TestIndex',
                'pk_template': '{id}',
                'sk_template': 'GSI',
            }
        ]
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'query_users_by_email',
                'description': 'Query by GSI',
                'operation': 'Query',
                'index_name': 'TestIndex',
                'consistent_read': True,  # Not allowed for GSI
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'entity_list',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        # Check that error message includes pattern_name
        assert any('query_users_by_email' in e.message for e in result.errors), (
            f"Expected 'query_users_by_email' in error messages, got: {[e.message for e in result.errors]}"
        )

    def test_consistent_read_gsi_error_explains_restriction(self, validator, valid_minimal_schema):
        """Test that GSI error message explains the restriction clearly."""
        # Add GSI to table config
        valid_minimal_schema['tables'][0]['gsi_list'] = [
            {
                'name': 'TestIndex',
                'partition_key': 'gsi_pk',
                'sort_key': 'gsi_sk',
            }
        ]
        # Add GSI mapping to entity
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['gsi_mappings'] = [
            {
                'name': 'TestIndex',
                'pk_template': '{id}',
                'sk_template': 'GSI',
            }
        ]
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'query_by_gsi',
                'description': 'Query by GSI',
                'operation': 'Query',
                'index_name': 'TestIndex',
                'consistent_read': True,  # Not allowed for GSI
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'entity_list',
            }
        ]
        result = self._validate_schema_dict(validator, valid_minimal_schema)
        assert not result.is_valid
        # Check that error message explains the GSI restriction
        assert any(
            'consistent_read cannot be true for GSI queries' in e.message
            and 'Global Secondary Index' in e.message
            and 'eventually consistent reads' in e.message
            for e in result.errors
        ), (
            f'Expected GSI restriction explanation in error messages, got: {[e.message for e in result.errors]}'
        )

    @given(
        consistent_read=st.one_of(
            st.text(),
            st.integers(),
            st.floats(allow_nan=False, allow_infinity=False),
            st.lists(st.booleans()),
            st.dictionaries(st.text(), st.booleans()),
            st.none(),
        )
    )
    def test_property_consistent_read_non_boolean_rejected(self, consistent_read):
        """Non-boolean values rejected.

        Test that for any non-boolean value, the schema validator rejects
        the pattern with a type validation error.
        """
        # Create validator and schema without using fixtures
        validator = SchemaValidator()

        valid_minimal_schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                        'sort_key': 'sk',
                    },
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'sk_template': 'ENTITY',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        # Create access pattern with non-boolean consistent_read value
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'test_pattern',
                'description': 'Test pattern',
                'operation': 'GetItem',
                'consistent_read': consistent_read,
                'parameters': [{'name': 'id', 'type': 'string'}],
                'return_type': 'single_entity',
            }
        ]

        result = self._validate_schema_dict(validator, valid_minimal_schema)

        # Should fail validation with boolean type error
        assert not result.is_valid
        assert any('consistent_read must be a boolean' in e.message.lower() for e in result.errors)

    @given(
        operation=st.sampled_from(['Query', 'Scan']),
        index_name=st.text(min_size=1, max_size=50).filter(lambda x: x.strip() != ''),
    )
    def test_property_gsi_queries_reject_consistent_read_true(self, operation, index_name):
        """Property Test 5: GSI queries reject consistent_read true.

        Test that for any GSI query (with index_name) and consistent_read: true,
        the schema validator rejects the pattern with a validation error explaining
        that GSIs do not support strongly consistent reads.
        """
        # Create validator and schema without using fixtures
        validator = SchemaValidator()

        valid_minimal_schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                        'sort_key': 'sk',
                    },
                    'gsi_list': [
                        {
                            'name': index_name,
                            'partition_key': 'gsi_pk',
                            'sort_key': 'gsi_sk',
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'sk_template': 'ENTITY',
                            'gsi_mappings': [
                                {
                                    'name': index_name,
                                    'pk_template': '{gsi_pk}',
                                    'sk_template': '{gsi_sk}',
                                }
                            ],
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                {'name': 'gsi_pk', 'type': 'string', 'required': True},
                                {'name': 'gsi_sk', 'type': 'string', 'required': True},
                            ],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        # Create GSI query access pattern with consistent_read: true
        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            {
                'pattern_id': 1,
                'name': 'test_gsi_query',
                'description': 'Test GSI query',
                'operation': operation,
                'index_name': index_name,
                'consistent_read': True,  # Not allowed for GSI
                'parameters': [{'name': 'gsi_pk', 'type': 'string'}],
                'return_type': 'entity_list',
            }
        ]

        result = self._validate_schema_dict(validator, valid_minimal_schema)

        # Should fail validation with GSI restriction error
        assert not result.is_valid
        assert any(
            'consistent_read cannot be true for GSI queries' in e.message
            and 'Global Secondary Index' in e.message
            for e in result.errors
        )

    @given(
        operation=st.sampled_from(['Query', 'Scan']),
        index_name=st.text(min_size=1, max_size=50).filter(lambda x: x.strip() != ''),
        consistent_read=st.one_of(st.just(False), st.none()),
    )
    def test_property_gsi_queries_accept_consistent_read_false_or_omitted(
        self, operation, index_name, consistent_read
    ):
        """GSI queries accept consistent_read false or omitted.

        consistent-read-parameter, Property 6: GSI queries accept consistent_read false or omitted

        Test that for any GSI query (with index_name), if consistent_read is set to false
        or omitted entirely, the schema validator accepts the pattern as valid.
        """
        # Create validator and schema without using fixtures
        validator = SchemaValidator()

        valid_minimal_schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                        'sort_key': 'sk',
                    },
                    'gsi_list': [
                        {
                            'name': index_name,
                            'partition_key': 'gsi_pk',
                            'sort_key': 'gsi_sk',
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'sk_template': 'ENTITY',
                            'gsi_mappings': [
                                {
                                    'name': index_name,
                                    'pk_template': '{gsi_pk}',
                                    'sk_template': '{gsi_sk}',
                                }
                            ],
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                {'name': 'gsi_pk', 'type': 'string', 'required': True},
                                {'name': 'gsi_sk', 'type': 'string', 'required': True},
                            ],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        # Create GSI query access pattern
        access_pattern = {
            'pattern_id': 1,
            'name': 'test_gsi_query',
            'description': 'Test GSI query',
            'operation': operation,
            'index_name': index_name,
            'parameters': [{'name': 'gsi_pk', 'type': 'string'}],
            'return_type': 'entity_list',
        }

        # Add consistent_read only if not None (to test omitted case)
        if consistent_read is not None:
            access_pattern['consistent_read'] = consistent_read

        valid_minimal_schema['tables'][0]['entities']['TestEntity']['access_patterns'] = [
            access_pattern
        ]

        result = self._validate_schema_dict(validator, valid_minimal_schema)

        # Should pass validation
        assert result.is_valid, (
            f'Expected valid schema but got errors: {[e.message for e in result.errors]}'
        )

    def test_validate_file_permission_error(self, validator):
        """Test handling of file permission errors (line 92)."""
        # Use a path that would trigger a permission/format error
        result = validator.validate_schema_file('/dev/null/nonexistent.json')
        assert not result.is_valid
        assert any('file' in e.path for e in result.errors)

    def test_validate_missing_pk_template_with_guidance(self, validator):
        """Test that missing pk_template provides template-specific guidance (lines 180-181)."""
        schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                    },
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            # Missing pk_template
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        assert not result.is_valid
        # Check for template-specific guidance
        assert any(
            'template syntax' in e.suggestion.lower() and 'USER#{user_id}' in e.suggestion
            for e in result.errors
        )

    def test_validate_array_field_with_item_type(self, validator):
        """Test that array fields with item_type are valid (lines 241-245)."""
        schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                    },
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                {
                                    'name': 'tags',
                                    'type': 'array',
                                    'item_type': 'string',
                                    'required': False,
                                },
                            ],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        # Should pass - array with item_type is valid
        assert result.is_valid

    def test_validate_main_table_range_query_exception_handling(self, validator):
        """Test exception handling in _validate_main_table_range_query (line 593)."""
        schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                    },
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [
                                {
                                    'pattern_id': 1,
                                    'name': 'test_range',
                                    'description': 'Test',
                                    'operation': 'Query',
                                    'range_condition': 'invalid_condition',  # Invalid
                                    'parameters': [{'name': 'id', 'type': 'string'}],
                                    'return_type': 'entity_list',
                                }
                            ],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        # Should have validation errors for invalid range condition
        assert not result.is_valid

    def test_validate_gsi_configuration_exception_handling(self, validator):
        """Test exception handling in _validate_gsi_configuration (lines 598-600)."""
        # Create a schema with malformed GSI structure that could trigger exceptions
        schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                    },
                    'gsi_list': 'not_a_list',  # Invalid type
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        # Should handle the error gracefully
        assert not result.is_valid

    def test_validate_file_value_error_non_json(self, validator):
        """Test handling of ValueError that's not JSON-related (line 92)."""
        # Create a file with invalid content that triggers ValueError but not JSON error
        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            f.write('')  # Empty file
            temp_file = f.name

        try:
            result = validator.validate_schema_file(temp_file)
            assert not result.is_valid
            # Should have json or file-related error
            assert any('json' in e.path or 'file' in e.path for e in result.errors)
        finally:
            os.unlink(temp_file)

    def test_validate_entity_type_wrong_type(self, validator):
        """Test that entity_type must be string (lines 241-242)."""
        schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                    },
                    'entities': {
                        'TestEntity': {
                            'entity_type': 123,  # Wrong type - should be string
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        assert not result.is_valid
        assert any('entity_type' in e.path for e in result.errors)

    def test_validate_range_query_with_range_errors(self, validator):
        """Test range query validation that produces errors (lines 556-557)."""
        schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                        'sort_key': 'sk',
                    },
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'sk_template': '{timestamp}',
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                {'name': 'timestamp', 'type': 'string', 'required': True},
                            ],
                            'access_patterns': [
                                {
                                    'pattern_id': 1,
                                    'name': 'test_range',
                                    'description': 'Test',
                                    'operation': 'Query',
                                    'range_condition': 'between',
                                    # Wrong number of parameters for 'between' (needs 3: pk + 2 range)
                                    'parameters': [{'name': 'id', 'type': 'string'}],
                                    'return_type': 'entity_list',
                                }
                            ],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        # Should have validation errors for wrong parameter count
        assert not result.is_valid
        assert any('parameter' in e.message.lower() for e in result.errors)

    def test_validate_range_query_exception_with_malformed_pattern(self, validator):
        """Test exception handling in range query validation (line 593)."""
        schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                    },
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [
                                {
                                    'pattern_id': 1,
                                    'name': 'test_range',
                                    'description': 'Test',
                                    'operation': 'Query',
                                    'range_condition': '>',
                                    # Missing required fields to trigger exception
                                    'parameters': None,  # Invalid
                                    'return_type': 'entity_list',
                                }
                            ],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        # Should handle exception gracefully
        assert not result.is_valid

    def test_validate_gsi_with_exception_in_validation(self, validator):
        """Test exception handling in GSI validation (lines 598-600)."""
        schema = {
            'tables': [
                {
                    'table_config': {
                        'table_name': 'TestTable',
                        'partition_key': 'pk',
                    },
                    'gsi_list': [
                        {
                            'name': 'TestGSI',
                            'partition_key': 'gsi_pk',
                            # Malformed structure that might cause issues
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'gsi_mappings': [
                                {
                                    'name': 'TestGSI',
                                    'pk_template': '{gsi_field}',
                                }
                            ],
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                # Missing gsi_field referenced in template
                            ],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }
        result = self._validate_schema_dict(validator, schema)
        # Should have validation errors
        assert not result.is_valid



"""Unit tests for GSI projection support."""

import json
import tempfile
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.gsi_validator import (
    GSIValidator,
)
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.schema_definitions import (
    VALID_GSI_PROJECTION_TYPES,
    GSIProjectionType,
)
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.schema_validator import (
    validate_schema_file,
)
from pathlib import Path


class TestGSIProjectionEnum:
    """Test GSI projection type enum."""

    def test_projection_types_exist(self):
        """Test all projection types are defined."""
        assert GSIProjectionType.ALL.value == 'ALL'
        assert GSIProjectionType.KEYS_ONLY.value == 'KEYS_ONLY'
        assert GSIProjectionType.INCLUDE.value == 'INCLUDE'

    def test_valid_projection_types_constant(self):
        """Test VALID_GSI_PROJECTION_TYPES contains all types."""
        assert 'ALL' in VALID_GSI_PROJECTION_TYPES
        assert 'KEYS_ONLY' in VALID_GSI_PROJECTION_TYPES
        assert 'INCLUDE' in VALID_GSI_PROJECTION_TYPES
        assert len(VALID_GSI_PROJECTION_TYPES) == 3


class TestGSIProjectionValidation:
    """Test GSI projection validation."""

    def test_valid_all_projection(self):
        """Test ALL projection validates successfully."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {'name': 'TestGSI', 'partition_key': 'gsi_pk', 'projection': 'ALL'}
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            assert result.is_valid
            assert len(result.errors) == 0
        finally:
            Path(temp_path).unlink()

    def test_valid_keys_only_projection(self):
        """Test KEYS_ONLY projection validates successfully."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {
                            'name': 'TestGSI',
                            'partition_key': 'gsi_pk',
                            'projection': 'KEYS_ONLY',
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            assert result.is_valid
            assert len(result.errors) == 0
        finally:
            Path(temp_path).unlink()

    def test_valid_include_projection(self):
        """Test INCLUDE projection with included_attributes validates."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {
                            'name': 'TestGSI',
                            'partition_key': 'gsi_pk',
                            'projection': 'INCLUDE',
                            'included_attributes': ['field1', 'field2'],
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                {'name': 'field1', 'type': 'string', 'required': True},
                                {'name': 'field2', 'type': 'string', 'required': True},
                            ],
                            'gsi_mappings': [{'name': 'TestGSI', 'pk_template': '{id}'}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            assert result.is_valid
            assert len(result.errors) == 0
        finally:
            Path(temp_path).unlink()

    def test_invalid_projection_type(self):
        """Test invalid projection type is rejected."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {
                            'name': 'TestGSI',
                            'partition_key': 'gsi_pk',
                            'projection': 'INVALID',
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            assert not result.is_valid
            assert any('invalid projection' in e.message.lower() for e in result.errors)
        finally:
            Path(temp_path).unlink()

    def test_include_without_attributes(self):
        """Test INCLUDE projection requires included_attributes."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {'name': 'TestGSI', 'partition_key': 'gsi_pk', 'projection': 'INCLUDE'}
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            assert not result.is_valid
            assert any('included_attributes' in e.message.lower() for e in result.errors)
        finally:
            Path(temp_path).unlink()

    def test_keys_only_with_attributes(self):
        """Test KEYS_ONLY rejects included_attributes."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {
                            'name': 'TestGSI',
                            'partition_key': 'gsi_pk',
                            'projection': 'KEYS_ONLY',
                            'included_attributes': ['field1'],
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [{'name': 'id', 'type': 'string', 'required': True}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            assert not result.is_valid
            assert any(
                'included_attributes' in e.message.lower()
                and 'only allowed for include' in e.message.lower()
                for e in result.errors
            )
        finally:
            Path(temp_path).unlink()

    def test_invalid_included_attribute(self):
        """Test included_attributes must reference valid fields."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {
                            'name': 'TestGSI',
                            'partition_key': 'gsi_pk',
                            'projection': 'INCLUDE',
                            'included_attributes': ['invalid_field'],
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                {'name': 'valid_field', 'type': 'string', 'required': True},
                            ],
                            'gsi_mappings': [{'name': 'TestGSI', 'pk_template': '{id}'}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            assert not result.is_valid
            assert any(
                'invalid_field' in e.message and 'not found' in e.message.lower()
                for e in result.errors
            )
        finally:
            Path(temp_path).unlink()

    def test_default_projection_is_all(self):
        """Test projection defaults to ALL when not specified."""
        validator = GSIValidator()
        table_data = {
            'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
            'gsi_list': [{'name': 'TestGSI', 'partition_key': 'gsi_pk'}],
            'entities': {},
        }

        gsi_list, errors = validator._parse_gsi_list(table_data, 'table')
        assert len(errors) == 0
        assert len(gsi_list) == 1
        assert gsi_list[0].projection == 'ALL'

    def test_projection_loaded_correctly(self):
        """Test projection field is loaded from schema."""
        validator = GSIValidator()
        table_data = {
            'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
            'gsi_list': [
                {'name': 'TestGSI', 'partition_key': 'gsi_pk', 'projection': 'KEYS_ONLY'}
            ],
            'entities': {},
        }

        gsi_list, errors = validator._parse_gsi_list(table_data, 'table')
        assert len(errors) == 0
        assert len(gsi_list) == 1
        assert gsi_list[0].projection == 'KEYS_ONLY'

    def test_included_attributes_loaded(self):
        """Test included_attributes are loaded from schema."""
        validator = GSIValidator()
        table_data = {
            'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
            'gsi_list': [
                {
                    'name': 'TestGSI',
                    'partition_key': 'gsi_pk',
                    'projection': 'INCLUDE',
                    'included_attributes': ['field1', 'field2'],
                }
            ],
            'entities': {},
        }

        gsi_list, errors = validator._parse_gsi_list(table_data, 'table')
        assert len(errors) == 0
        assert len(gsi_list) == 1
        assert gsi_list[0].included_attributes == ['field1', 'field2']

    def test_smart_warning_for_required_non_projected_fields(self):
        """Test smart validation warns when INCLUDE has required non-projected fields."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {
                            'name': 'TestGSI',
                            'partition_key': 'gsi_pk',
                            'projection': 'INCLUDE',
                            'included_attributes': ['field1'],
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                {'name': 'field1', 'type': 'string', 'required': True},
                                {
                                    'name': 'field2',
                                    'type': 'string',
                                    'required': True,
                                },  # Required but NOT projected
                            ],
                            'gsi_mappings': [{'name': 'TestGSI', 'pk_template': '{id}'}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            # Should be valid but have warnings
            assert result.is_valid
            assert len(result.warnings) > 0
            assert any('field2' in w.message for w in result.warnings)
            assert any(
                'required fields not in included_attributes' in w.message for w in result.warnings
            )
        finally:
            Path(temp_path).unlink()

    def test_no_warning_when_non_projected_fields_optional(self):
        """Test no warning when all non-projected fields are optional."""
        schema = {
            'tables': [
                {
                    'table_config': {'table_name': 'Test', 'partition_key': 'pk'},
                    'gsi_list': [
                        {
                            'name': 'TestGSI',
                            'partition_key': 'gsi_pk',
                            'projection': 'INCLUDE',
                            'included_attributes': ['field1'],
                        }
                    ],
                    'entities': {
                        'TestEntity': {
                            'entity_type': 'TEST',
                            'pk_template': '{id}',
                            'fields': [
                                {'name': 'id', 'type': 'string', 'required': True},
                                {'name': 'field1', 'type': 'string', 'required': True},
                                {
                                    'name': 'field2',
                                    'type': 'string',
                                    'required': False,
                                },  # Optional - safe!
                            ],
                            'gsi_mappings': [{'name': 'TestGSI', 'pk_template': '{id}'}],
                            'access_patterns': [],
                        }
                    },
                }
            ]
        }

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(schema, f)
            temp_path = f.name

        try:
            result = validate_schema_file(temp_path)
            # Should be valid with no warnings
            assert result.is_valid
            assert len(result.warnings) == 0
        finally:
            Path(temp_path).unlink()



# Auto-generated entities
from __future__ import annotations

from base_repository import ConfigurableEntity, EntityConfig
from decimal import Decimal
from typing import Any


# Organization Entity Configuration
ORGANIZATION_CONFIG = EntityConfig(
    entity_type='ORGANIZATION',
    pk_builder=lambda entity: f'ORG#{entity.org_id}',
    pk_lookup_builder=lambda org_id: f'ORG#{org_id}',
    sk_builder=lambda entity: 'DETAILS',
    sk_lookup_builder=lambda: 'DETAILS',
    prefix_builder=lambda **kwargs: 'ORGANIZATION#',
)


class Organization(ConfigurableEntity):
    org_id: str
    name: str
    domain: str
    plan_type: str
    max_users: int
    max_projects: int
    created_at: str
    updated_at: str
    status: str
    billing_email: str
    settings: dict[str, Any] = None

    @classmethod
    def get_config(cls) -> EntityConfig:
        return ORGANIZATION_CONFIG


# OrganizationInvite Entity Configuration
ORGANIZATIONINVITE_CONFIG = EntityConfig(
    entity_type='INVITE',
    pk_builder=lambda entity: f'ORG#{entity.org_id}',
    pk_lookup_builder=lambda org_id: f'ORG#{org_id}',
    sk_builder=lambda entity: f'INVITE#{entity.invite_id}',
    sk_lookup_builder=lambda invite_id: f'INVITE#{invite_id}',
    prefix_builder=lambda **kwargs: 'INVITE#',
)


class OrganizationInvite(ConfigurableEntity):
    org_id: str
    invite_id: str
    email: str
    role: str
    invited_by: str
    created_at: str
    expires_at: str
    status: str
    accepted_at: str = None

    @classmethod
    def get_config(cls) -> EntityConfig:
        return ORGANIZATIONINVITE_CONFIG


# OrganizationMember Entity Configuration
ORGANIZATIONMEMBER_CONFIG = EntityConfig(
    entity_type='MEMBER',
    pk_builder=lambda entity: f'ORG#{entity.org_id}',
    pk_lookup_builder=lambda org_id: f'ORG#{org_id}',
    sk_builder=lambda entity: f'MEMBER#{entity.user_id}',
    sk_lookup_builder=lambda user_id: f'MEMBER#{user_id}',
    prefix_builder=lambda **kwargs: 'MEMBER#',
)


class OrganizationMember(ConfigurableEntity):
    org_id: str
    user_id: str
    email: str
    first_name: str
    last_name: str
    role: str
    permissions: list[str]
    joined_at: str
    last_active: str = None
    status: str

    @classmethod
    def get_config(cls) -> EntityConfig:
        return ORGANIZATIONMEMBER_CONFIG


# OrganizationProject Entity Configuration
ORGANIZATIONPROJECT_CONFIG = EntityConfig(
    entity_type='ORG_PROJECT',
    pk_builder=lambda entity: f'ORG#{entity.org_id}',
    pk_lookup_builder=lambda org_id: f'ORG#{org_id}',
    sk_builder=lambda entity: f'PROJECT#{entity.created_at}#{entity.project_id}',
    sk_lookup_builder=lambda created_at, project_id: f'PROJECT#{created_at}#{project_id}',
    prefix_builder=lambda **kwargs: 'PROJECT#',
)


class OrganizationProject(ConfigurableEntity):
    org_id: str
    project_id: str
    project_name: str
    status: str
    priority: str
    owner_id: str
    team_size: int
    created_at: str
    due_date: str = None

    @classmethod
    def get_config(cls) -> EntityConfig:
        return ORGANIZATIONPROJECT_CONFIG


# Project Entity Configuration
PROJECT_CONFIG = EntityConfig(
    entity_type='PROJECT',
    pk_builder=lambda entity: f'PROJECT#{entity.project_id}',
    pk_lookup_builder=lambda project_id: f'PROJECT#{project_id}',
    sk_builder=lambda entity: 'DETAILS',
    sk_lookup_builder=lambda: 'DETAILS',
    prefix_builder=lambda **kwargs: 'PROJECT#',
)


class Project(ConfigurableEntity):
    project_id: str
    org_id: str
    name: str
    description: str = None
    status: str
    priority: str
    owner_id: str
    team_members: list[str]
    start_date: str = None
    due_date: str = None
    budget: Decimal = None
    currency: str = None
    tags: list[str] = None
    created_at: str
    updated_at: str

    @classmethod
    def get_config(cls) -> EntityConfig:
        return PROJECT_CONFIG


# ProjectMilestone Entity Configuration
PROJECTMILESTONE_CONFIG = EntityConfig(
    entity_type='MILESTONE',
    pk_builder=lambda entity: f'PROJECT#{entity.project_id}',
    pk_lookup_builder=lambda project_id: f'PROJECT#{project_id}',
    sk_builder=lambda entity: f'MILESTONE#{entity.milestone_id}',
    sk_lookup_builder=lambda milestone_id: f'MILESTONE#{milestone_id}',
    prefix_builder=lambda **kwargs: 'MILESTONE#',
)


class ProjectMilestone(ConfigurableEntity):
    project_id: str
    milestone_id: str
    title: str
    description: str = None
    due_date: str
    status: str
    completion_percentage: int
    created_at: str
    completed_at: str = None

    @classmethod
    def get_config(cls) -> EntityConfig:
        return PROJECTMILESTONE_CONFIG


# ProjectTask Entity Configuration
PROJECTTASK_CONFIG = EntityConfig(
    entity_type='PROJECT_TASK',
    pk_builder=lambda entity: f'PROJECT#{entity.project_id}',
    pk_lookup_builder=lambda project_id: f'PROJECT#{project_id}',
    sk_builder=lambda entity: f'TASK#{entity.status}#{entity.priority}#{entity.task_id}',
    sk_lookup_builder=lambda status, priority, task_id: f'TASK#{status}#{priority}#{task_id}',
    prefix_builder=lambda **kwargs: 'TASK#',
)


class ProjectTask(ConfigurableEntity):
    project_id: str
    task_id: str
    title: str
    status: str
    priority: str
    assignee_id: str = None
    due_date: str = None
    estimated_hours: Decimal = None
    created_at: str

    @classmethod
    def get_config(cls) -> EntityConfig:
        return PROJECTTASK_CONFIG


# Task Entity Configuration
TASK_CONFIG = EntityConfig(
    entity_type='TASK',
    pk_builder=lambda entity: f'TASK#{entity.task_id}',
    pk_lookup_builder=lambda task_id: f'TASK#{task_id}',
    sk_builder=lambda entity: 'DETAILS',
    sk_lookup_builder=lambda: 'DETAILS',
    prefix_builder=lambda **kwargs: 'TASK#',
)


class Task(ConfigurableEntity):
    task_id: str
    project_id: str
    title: str
    description: str = None
    status: str
    priority: str
    assignee_id: str = None
    reporter_id: str
    estimated_hours: Decimal = None
    actual_hours: Decimal = None
    due_date: str = None
    labels: list[str] = None
    dependencies: list[str] = None
    created_at: str
    updated_at: str
    completed_at: str = None

    @classmethod
    def get_config(cls) -> EntityConfig:
        return TASK_CONFIG


# TaskComment Entity Configuration
TASKCOMMENT_CONFIG = EntityConfig(
    entity_type='COMMENT',
    pk_builder=lambda entity: f'TASK#{entity.task_id}',
    pk_lookup_builder=lambda task_id: f'TASK#{task_id}',
    sk_builder=lambda entity: f'COMMENT#{entity.created_at}#{entity.comment_id}',
    sk_lookup_builder=lambda created_at, comment_id: f'COMMENT#{created_at}#{comment_id}',
    prefix_builder=lambda **kwargs: 'COMMENT#',
)


class TaskComment(ConfigurableEntity):
    task_id: str
    comment_id: str
    author_id: str
    content: str
    comment_type: str
    created_at: str
    updated_at: str = None
    mentions: list[str] = None
    attachments: list[str] = None

    @classmethod
    def get_config(cls) -> EntityConfig:
        return TASKCOMMENT_CONFIG


# UserTask Entity Configuration
USERTASK_CONFIG = EntityConfig(
    entity_type='USER_TASK',
    pk_builder=lambda entity: f'USER#{entity.user_id}',
    pk_lookup_builder=lambda user_id: f'USER#{user_id}',
    sk_builder=lambda entity: f'TASK#{entity.status}#{entity.due_date}#{entity.task_id}',
    sk_lookup_builder=lambda status, due_date, task_id: f'TASK#{status}#{due_date}#{task_id}',
    prefix_builder=lambda **kwargs: 'TASK#',
)


class UserTask(ConfigurableEntity):
    user_id: str
    task_id: str
    project_id: str
    title: str
    status: str
    priority: str
    due_date: str = None
    estimated_hours: Decimal = None
    assigned_at: str

    @classmethod
    def get_config(cls) -> EntityConfig:
        return USERTASK_CONFIG



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Output management for generated code files."""

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any


@dataclass
class GeneratedFile:
    """Represents a single generated file."""

    path: str  # Relative path: "entities.py" or "models/UserProfile.java"
    description: str  # Human description: "5 entities" or "UserProfile entity"
    category: str  # File category: "entities", "repositories", "config", "examples"
    content: str = ''  # Complete file content (for new approach)
    count: int = 0  # Optional count for summary


@dataclass
class GenerationResult:
    """Contains all generated code and metadata."""

    generated_files: list[GeneratedFile]  # List of all files created
    access_pattern_mapping: dict[str, Any]
    generator_type: str


class OutputManager:
    """Manages all output operations for generated code."""

    def __init__(self, output_dir: str, language: str = 'python'):
        """Initialize the output manager with target directory and language."""
        self.output_path = Path(output_dir)
        self.language = language
        self.output_path.mkdir(parents=True, exist_ok=True)

    def write_generated_files(self, generation_result: GenerationResult) -> None:
        """Write all generated files in one coordinated operation."""
        # Write files from manifest
        for file in generation_result.generated_files:
            if file.content:  # Files with content - write directly
                self._write_file(file.path, file.content)
            else:  # Files without content - copy from source
                self._copy_support_file(file.path)

        # Always write the access pattern mapping
        self._write_mapping_file(
            generation_result.access_pattern_mapping, generation_result.generator_type
        )

        self._print_summary(generation_result)

    def _write_mapping_file(
        self, access_pattern_mapping: dict[str, Any], generator_type: str
    ) -> None:
        """Write access_pattern_mapping.json."""
        mapping_file = self.output_path / 'access_pattern_mapping.json'
        with open(mapping_file, 'w') as f:
            json.dump(
                {
                    'metadata': {
                        'generated_at': {'timestamp': 'auto-generated'},
                        'total_patterns': len(access_pattern_mapping),
                        'generator_type': generator_type,
                    },
                    'access_pattern_mapping': access_pattern_mapping,
                },
                f,
                indent=2,
            )
            f.write('\n')  # Add trailing newline for pre-commit compatibility

    def _write_file(self, file_path: str, content: str) -> None:
        """Write a single file with content - language agnostic."""
        output_file = self.output_path / file_path

        # Create parent directories if they don't exist (for nested paths like "models/UserProfile.java")
        output_file.parent.mkdir(parents=True, exist_ok=True)

        with open(output_file, 'w') as f:
            f.write(content)
            if not content.endswith('\n'):
                f.write('\n')  # Add final newline

    def _copy_support_file(self, dest_filename: str) -> None:
        """Copy a single support file from language-specific directory."""
        # Get the generator directory (parent of parent of this file)
        generator_dir = Path(__file__).parent.parent
        language_dir = generator_dir / 'languages' / self.language

        # Copy the file
        source_file = language_dir / dest_filename
        dest_file = self.output_path / dest_filename

        if source_file.exists():
            # Create parent directories if they don't exist (for nested paths)
            dest_file.parent.mkdir(parents=True, exist_ok=True)

            with open(source_file) as src, open(dest_file, 'w') as dst:
                dst.write(src.read())
        else:
            print(f'Warning: Support file not found: {source_file}')

    def _print_summary(self, generation_result: GenerationResult) -> None:
        """Print generation summary with flexible file listing."""
        print(f'Generated code in {self.output_path}')

        # Group files by category for organized output
        by_category = {}
        for file in generation_result.generated_files:
            if file.category not in by_category:
                by_category[file.category] = []
            by_category[file.category].append(file)

        # Print organized summary in logical order
        for category in ['entities', 'repositories', 'config', 'examples']:
            if category in by_category:
                files = by_category[category]
                if len(files) == 1 and files[0].count > 0:
                    # Single file with count: "- entities.py: 5 entities"
                    print(f'- {files[0].path}: {files[0].description}')
                else:
                    # Multiple files or no count: list each file
                    for file in files:
                        print(f'- {file.path}: {file.description}')

        # Always show access pattern mapping
        print(
            f'- access_pattern_mapping.json: {len(generation_result.access_pattern_mapping)} access patterns'
        )



"""Generated usage examples for DynamoDB entities and repositories"""

from __future__ import annotations

import os
import sys

# Import generated entities and repositories
from entities import (
    TenantCertificate,
    TenantCourse,
    TenantEnrollment,
    TenantLesson,
    TenantOrganization,
    TenantProgress,
    TenantUser,
)
from repositories import (
    TenantCertificateRepository,
    TenantCourseRepository,
    TenantEnrollmentRepository,
    TenantLessonRepository,
    TenantOrganizationRepository,
    TenantProgressRepository,
    TenantUserRepository,
)


class UsageExamples:
    """Examples of using the generated entities and repositories"""

    def __init__(self):
        """Initialize repositories with default table names from schema."""
        # Initialize repositories with their respective table names
        # ELearningPlatform table repositories
        try:
            self.tenantcertificate_repo = TenantCertificateRepository('ELearningPlatform')
            print("✅ Initialized TenantCertificateRepository for table 'ELearningPlatform'")
        except Exception as e:
            print(f'❌ Failed to initialize TenantCertificateRepository: {e}')
            self.tenantcertificate_repo = None
        try:
            self.tenantcourse_repo = TenantCourseRepository('ELearningPlatform')
            print("✅ Initialized TenantCourseRepository for table 'ELearningPlatform'")
        except Exception as e:
            print(f'❌ Failed to initialize TenantCourseRepository: {e}')
            self.tenantcourse_repo = None
        try:
            self.tenantenrollment_repo = TenantEnrollmentRepository('ELearningPlatform')
            print("✅ Initialized TenantEnrollmentRepository for table 'ELearningPlatform'")
        except Exception as e:
            print(f'❌ Failed to initialize TenantEnrollmentRepository: {e}')
            self.tenantenrollment_repo = None
        try:
            self.tenantlesson_repo = TenantLessonRepository('ELearningPlatform')
            print("✅ Initialized TenantLessonRepository for table 'ELearningPlatform'")
        except Exception as e:
            print(f'❌ Failed to initialize TenantLessonRepository: {e}')
            self.tenantlesson_repo = None
        try:
            self.tenantorganization_repo = TenantOrganizationRepository('ELearningPlatform')
            print("✅ Initialized TenantOrganizationRepository for table 'ELearningPlatform'")
        except Exception as e:
            print(f'❌ Failed to initialize TenantOrganizationRepository: {e}')
            self.tenantorganization_repo = None
        try:
            self.tenantprogress_repo = TenantProgressRepository('ELearningPlatform')
            print("✅ Initialized TenantProgressRepository for table 'ELearningPlatform'")
        except Exception as e:
            print(f'❌ Failed to initialize TenantProgressRepository: {e}')
            self.tenantprogress_repo = None
        try:
            self.tenantuser_repo = TenantUserRepository('ELearningPlatform')
            print("✅ Initialized TenantUserRepository for table 'ELearningPlatform'")
        except Exception as e:
            print(f'❌ Failed to initialize TenantUserRepository: {e}')
            self.tenantuser_repo = None

    def run_examples(self, include_additional_access_patterns: bool = False):
        """Run CRUD examples for all entities"""
        # Dictionary to store created entities for access pattern testing
        created_entities = {}

        # Step 0: Cleanup any leftover entities from previous runs (makes tests idempotent)
        print('🧹 Pre-test Cleanup: Removing any leftover entities from previous runs')
        print('=' * 50)
        # Try to delete TenantCertificate (tenant_id, user_id, course_id, issued_date)
        try:
            sample_tenantcertificate = TenantCertificate(
                tenant_id='tenant-12345',
                user_id='user-67890',
                course_id='course-11111',
                certificate_id='cert-22222',
                course_title='JavaScript Fundamentals',
                user_name='John Doe',
                instructor_name='Jane Smith',
                issued_date=1705737600,
                completion_date=1705651200,
                final_grade='A',
                certificate_url='https://example.com/certificates/cert-22222.pdf',
                verification_code='VERIFY-12345',
                expiry_date=1737273600,
                status='active',
            )
            self.tenantcertificate_repo.delete_tenant_certificate(
                sample_tenantcertificate.tenant_id,
                sample_tenantcertificate.user_id,
                sample_tenantcertificate.course_id,
                sample_tenantcertificate.issued_date,
            )
            print('   🗑️  Deleted leftover tenantcertificate (if existed)')
        except Exception:
            pass  # Ignore errors - item might not exist
        # Try to delete TenantCourse (tenant_id, course_id)
        try:
            sample_tenantcourse = TenantCourse(
                tenant_id='tenant-12345',
                course_id='course-11111',
                title='JavaScript Fundamentals',
                description='Learn the basics of JavaScript programming language',
                instructor_id='instructor-001',
                instructor_name='John Smith',
                category='Programming',
                difficulty_level='beginner',
                duration_hours=40,
                max_enrollments=100,
                prerequisites=['Basic Computer Skills'],
                tags=['javascript', 'programming', 'web-development'],
                created_at=1704067200,
                updated_at=1705737600,
                status='active',
            )
            self.tenantcourse_repo.delete_tenant_course(
                sample_tenantcourse.tenant_id, sample_tenantcourse.course_id
            )
            print('   🗑️  Deleted leftover tenantcourse (if existed)')
        except Exception:
            pass  # Ignore errors - item might not exist
        # Try to delete TenantEnrollment (tenant_id, user_id, course_id, enrollment_date)
        try:
            sample_tenantenrollment = TenantEnrollment(
                tenant_id='tenant-12345',
                user_id='user-67890',
                course_id='course-11111',
                course_title='JavaScript Fundamentals',
                instructor_name='John Smith',
                enrollment_date=1705305600,
                completion_date=42,
                progress_percentage=75,
                current_lesson='lesson-33333',
                grade='None',
                certificate_issued=False,
                status='active',
            )
            self.tenantenrollment_repo.delete_tenant_enrollment(
                sample_tenantenrollment.tenant_id,
                sample_tenantenrollment.user_id,
                sample_tenantenrollment.course_id,
                sample_tenantenrollment.enrollment_date,
            )
            print('   🗑️  Deleted leftover tenantenrollment (if existed)')
        except Exception:
            pass  # Ignore errors - item might not exist
        # Try to delete TenantLesson (tenant_id, course_id, lesson_order, lesson_id)
        try:
            sample_tenantlesson = TenantLesson(
                tenant_id='tenant-12345',
                course_id='course-11111',
                lesson_id='lesson-33333',
                lesson_order=1,
                title='Introduction to Variables',
                description='Learn about JavaScript variables and data types',
                content_type='video',
                content_url='https://example.com/videos/lesson-33333.mp4',
                duration_minutes=25,
                is_mandatory=True,
                quiz_required=True,
                passing_score=80,
                created_at=1704067200,
                updated_at=1705737600,
            )
            self.tenantlesson_repo.delete_tenant_lesson(
                sample_tenantlesson.tenant_id,
                sample_tenantlesson.course_id,
                sample_tenantlesson.lesson_order,
                sample_tenantlesson.lesson_id,
            )
            print('   🗑️  Deleted leftover tenantlesson (if existed)')
        except Exception:
            pass  # Ignore errors - item might not exist
        # Try to delete TenantOrganization (tenant_id)
        try:
            sample_tenantorganization = TenantOrganization(
                tenant_id='tenant-12345',
                organization_name='TechCorp Learning',
                domain='techcorp.com',
                subscription_plan='enterprise',
                max_users=500,
                max_courses=100,
                admin_email='admin@techcorp.com',
                created_at=1704067200,
                status='active',
            )
            self.tenantorganization_repo.delete_tenant_organization(
                sample_tenantorganization.tenant_id
            )
            print('   🗑️  Deleted leftover tenantorganization (if existed)')
        except Exception:
            pass  # Ignore errors - item might not exist
        # Try to delete TenantProgress (tenant_id, user_id, course_id, lesson_id, attempt_date)
        try:
            sample_tenantprogress = TenantProgress(
                tenant_id='tenant-12345',
                user_id='user-67890',
                course_id='course-11111',
                lesson_id='lesson-33333',
                attempt_date=1705737600,
                completion_status='completed',
                time_spent_minutes=30,
                quiz_score=92,
                quiz_passed=True,
                notes='Great progress on variables concept',
                last_accessed=1705824000,
            )
            self.tenantprogress_repo.delete_tenant_progress(
                sample_tenantprogress.tenant_id,
                sample_tenantprogress.user_id,
                sample_tenantprogress.course_id,
                sample_tenantprogress.lesson_id,
                sample_tenantprogress.attempt_date,
            )
            print('   🗑️  Deleted leftover tenantprogress (if existed)')
        except Exception:
            pass  # Ignore errors - item might not exist
        # Try to delete TenantUser (tenant_id, user_id)
        try:
            sample_tenantuser = TenantUser(
                tenant_id='tenant-12345',
                user_id='user-67890',
                email='john.doe@techcorp.com',
                first_name='John',
                last_name='Doe',
                role='learner',
                department='Engineering',
                job_title='Software Developer',
                enrollment_date=1704931200,
                last_login=1705737600,
                status='active',
            )
            self.tenantuser_repo.delete_tenant_user(
                sample_tenantuser.tenant_id, sample_tenantuser.user_id
            )
            print('   🗑️  Deleted leftover tenantuser (if existed)')
        except Exception:
            pass  # Ignore errors - item might not exist
        print('✅ Pre-test cleanup completed\n')

        print('Running Repository Examples')
        print('=' * 50)
        print('\n=== ELearningPlatform Table Operations ===')

        # TenantCertificate example
        print('\n--- TenantCertificate ---')

        # 1. CREATE - Create sample tenantcertificate
        sample_tenantcertificate = TenantCertificate(
            tenant_id='tenant-12345',
            user_id='user-67890',
            course_id='course-11111',
            certificate_id='cert-22222',
            course_title='JavaScript Fundamentals',
            user_name='John Doe',
            instructor_name='Jane Smith',
            issued_date=1705737600,
            completion_date=1705651200,
            final_grade='A',
            certificate_url='https://example.com/certificates/cert-22222.pdf',
            verification_code='VERIFY-12345',
            expiry_date=1737273600,
            status='active',
        )

        print('📝 Creating tenantcertificate...')
        print(f'📝 PK: {sample_tenantcertificate.pk()}, SK: {sample_tenantcertificate.sk()}')

        try:
            created_tenantcertificate = self.tenantcertificate_repo.create_tenant_certificate(
                sample_tenantcertificate
            )
            print(f'✅ Created: {created_tenantcertificate}')
            # Store created entity for access pattern testing
            created_entities['TenantCertificate'] = created_tenantcertificate
        except Exception as e:
            # Check if the error is due to item already existing
            if 'ConditionalCheckFailedException' in str(e) or 'already exists' in str(e).lower():
                print('⚠️  tenantcertificate already exists, retrieving existing entity...')
                try:
                    existing_tenantcertificate = (
                        self.tenantcertificate_repo.get_tenant_certificate(
                            sample_tenantcertificate.tenant_id,
                            sample_tenantcertificate.user_id,
                            sample_tenantcertificate.course_id,
                            sample_tenantcertificate.issued_date,
                        )
                    )

                    if existing_tenantcertificate:
                        print(f'✅ Retrieved existing: {existing_tenantcertificate}')
                        # Store existing entity for access pattern testing
                        created_entities['TenantCertificate'] = existing_tenantcertificate
                    else:
                        print('❌ Failed to retrieve existing tenantcertificate')
                except Exception as get_error:
                    print(f'❌ Failed to retrieve existing tenantcertificate: {get_error}')
            else:
                print(f'❌ Failed to create tenantcertificate: {e}')
        # 2. UPDATE - Update non-key field (course_title)
        if 'TenantCertificate' in created_entities:
            print('\n🔄 Updating course_title field...')
            try:
                # Refresh entity to get latest version (handles optimistic locking)
                entity_for_refresh = created_entities['TenantCertificate']
                refreshed_entity = self.tenantcertificate_repo.get_tenant_certificate(
                    entity_for_refresh.tenant_id,
                    entity_for_refresh.user_id,
                    entity_for_refresh.course_id,
                    entity_for_refresh.issued_date,
                )

                if refreshed_entity:
                    original_value = refreshed_entity.course_title
                    refreshed_entity.course_title = 'Advanced JavaScript Fundamentals'

                    updated_tenantcertificate = (
                        self.tenantcertificate_repo.update_tenant_certificate(refreshed_entity)
                    )
                    print(
                        f'✅ Updated course_title: {original_value} → {updated_tenantcertificate.course_title}'
                    )

                    # Update stored entity with updated values
                    created_entities['TenantCertificate'] = updated_tenantcertificate
                else:
                    print('❌ Could not refresh tenantcertificate for update')
            except Exception as e:
                if 'version' in str(e).lower() or 'modified by another process' in str(e).lower():
                    print(
                        f'⚠️  tenantcertificate was modified by another process (optimistic locking): {e}'
                    )
                    print('💡 This is expected behavior in concurrent environments')
                else:
                    print(f'❌ Failed to update tenantcertificate: {e}')

        # 3. GET - Retrieve and print the entity
        if 'TenantCertificate' in created_entities:
            print('\n🔍 Retrieving tenantcertificate...')
            try:
                entity_for_get = created_entities['TenantCertificate']
                retrieved_tenantcertificate = self.tenantcertificate_repo.get_tenant_certificate(
                    entity_for_get.tenant_id,
                    entity_for_get.user_id,
                    entity_for_get.course_id,
                    entity_for_get.issued_date,
                )

                if retrieved_tenantcertificate:
                    print(f'✅ Retrieved: {retrieved_tenantcertificate}')
                else:
                    print('❌ Failed to retrieve tenantcertificate')
            except Exception as e:
                print(f'❌ Failed to retrieve tenantcertificate: {e}')

        print('🎯 TenantCertificate CRUD cycle completed!')

        # TenantCourse example
        print('\n--- TenantCourse ---')

        # 1. CREATE - Create sample tenantcourse
        sample_tenantcourse = TenantCourse(
            tenant_id='tenant-12345',
            course_id='course-11111',
            title='JavaScript Fundamentals',
            description='Learn the basics of JavaScript programming language',
            instructor_id='instructor-001',
            instructor_name='John Smith',
            category='Programming',
            difficulty_level='beginner',
            duration_hours=40,
            max_enrollments=100,
            prerequisites=['Basic Computer Skills'],
            tags=['javascript', 'programming', 'web-development'],
            created_at=1704067200,
            updated_at=1705737600,
            status='active',
        )

        print('📝 Creating tenantcourse...')
        print(f'📝 PK: {sample_tenantcourse.pk()}, SK: {sample_tenantcourse.sk()}')

        try:
            created_tenantcourse = self.tenantcourse_repo.create_tenant_course(sample_tenantcourse)
            print(f'✅ Created: {created_tenantcourse}')
            # Store created entity for access pattern testing
            created_entities['TenantCourse'] = created_tenantcourse
        except Exception as e:
            # Check if the error is due to item already existing
            if 'ConditionalCheckFailedException' in str(e) or 'already exists' in str(e).lower():
                print('⚠️  tenantcourse already exists, retrieving existing entity...')
                try:
                    existing_tenantcourse = self.tenantcourse_repo.get_tenant_course(
                        sample_tenantcourse.tenant_id, sample_tenantcourse.course_id
                    )

                    if existing_tenantcourse:
                        print(f'✅ Retrieved existing: {existing_tenantcourse}')
                        # Store existing entity for access pattern testing
                        created_entities['TenantCourse'] = existing_tenantcourse
                    else:
                        print('❌ Failed to retrieve existing tenantcourse')
                except Exception as get_error:
                    print(f'❌ Failed to retrieve existing tenantcourse: {get_error}')
            else:
                print(f'❌ Failed to create tenantcourse: {e}')
        # 2. UPDATE - Update non-key field (title)
        if 'TenantCourse' in created_entities:
            print('\n🔄 Updating title field...')
            try:
                # Refresh entity to get latest version (handles optimistic locking)
                entity_for_refresh = created_entities['TenantCourse']
                refreshed_entity = self.tenantcourse_repo.get_tenant_course(
                    entity_for_refresh.tenant_id, entity_for_refresh.course_id
                )

                if refreshed_entity:
                    original_value = refreshed_entity.title
                    refreshed_entity.title = 'Advanced JavaScript Fundamentals'

                    updated_tenantcourse = self.tenantcourse_repo.update_tenant_course(
                        refreshed_entity
                    )
                    print(f'✅ Updated title: {original_value} → {updated_tenantcourse.title}')

                    # Update stored entity with updated values
                    created_entities['TenantCourse'] = updated_tenantcourse
                else:
                    print('❌ Could not refresh tenantcourse for update')
            except Exception as e:
                if 'version' in str(e).lower() or 'modified by another process' in str(e).lower():
                    print(
                        f'⚠️  tenantcourse was modified by another process (optimistic locking): {e}'
                    )
                    print('💡 This is expected behavior in concurrent environments')
                else:
                    print(f'❌ Failed to update tenantcourse: {e}')

        # 3. GET - Retrieve and print the entity
        if 'TenantCourse' in created_entities:
            print('\n🔍 Retrieving tenantcourse...')
            try:
                entity_for_get = created_entities['TenantCourse']
                retrieved_tenantcourse = self.tenantcourse_repo.get_tenant_course(
                    entity_for_get.tenant_id, entity_for_get.course_id
                )

                if retrieved_tenantcourse:
                    print(f'✅ Retrieved: {retrieved_tenantcourse}')
                else:
                    print('❌ Failed to retrieve tenantcourse')
            except Exception as e:
                print(f'❌ Failed to retrieve tenantcourse: {e}')

        print('🎯 TenantCourse CRUD cycle completed!')

        # TenantEnrollment example
        print('\n--- TenantEnrollment ---')

        # 1. CREATE - Create sample tenantenrollment
        sample_tenantenrollment = TenantEnrollment(
            tenant_id='tenant-12345',
            user_id='user-67890',
            course_id='course-11111',
            course_title='JavaScript Fundamentals',
            instructor_name='John Smith',
            enrollment_date=1705305600,
            completion_date=42,
            progress_percentage=75,
            current_lesson='lesson-33333',
            grade='None',
            certificate_issued=False,
            status='active',
        )

        print('📝 Creating tenantenrollment...')
        print(f'📝 PK: {sample_tenantenrollment.pk()}, SK: {sample_tenantenrollment.sk()}')

        try:
            created_tenantenrollment = self.tenantenrollment_repo.create_tenant_enrollment(
                sample_tenantenrollment
            )
            print(f'✅ Created: {created_tenantenrollment}')
            # Store created entity for access pattern testing
            created_entities['TenantEnrollment'] = created_tenantenrollment
        except Exception as e:
            # Check if the error is due to item already existing
            if 'ConditionalCheckFailedException' in str(e) or 'already exists' in str(e).lower():
                print('⚠️  tenantenrollment already exists, retrieving existing entity...')
                try:
                    existing_tenantenrollment = self.tenantenrollment_repo.get_tenant_enrollment(
                        sample_tenantenrollment.tenant_id,
                        sample_tenantenrollment.user_id,
                        sample_tenantenrollment.course_id,
                        sample_tenantenrollment.enrollment_date,
                    )

                    if existing_tenantenrollment:
                        print(f'✅ Retrieved existing: {existing_tenantenrollment}')
                        # Store existing entity for access pattern testing
                        created_entities['TenantEnrollment'] = existing_tenantenrollment
                    else:
                        print('❌ Failed to retrieve existing tenantenrollment')
                except Exception as get_error:
                    print(f'❌ Failed to retrieve existing tenantenrollment: {get_error}')
            else:
                print(f'❌ Failed to create tenantenrollment: {e}')
        # 2. UPDATE - Update non-key field (completion_date)
        if 'TenantEnrollment' in created_entities:
            print('\n🔄 Updating completion_date field...')
            try:
                # Refresh entity to get latest version (handles optimistic locking)
                entity_for_refresh = created_entities['TenantEnrollment']
                refreshed_entity = self.tenantenrollment_repo.get_tenant_enrollment(
                    entity_for_refresh.tenant_id,
                    entity_for_refresh.user_id,
                    entity_for_refresh.course_id,
                    entity_for_refresh.enrollment_date,
                )

                if refreshed_entity:
                    original_value = refreshed_entity.completion_date
                    refreshed_entity.completion_date = 1705824000

                    updated_tenantenrollment = self.tenantenrollment_repo.update_tenant_enrollment(
                        refreshed_entity
                    )
                    print(
                        f'✅ Updated completion_date: {original_value} → {updated_tenantenrollment.completion_date}'
                    )

                    # Update stored entity with updated values
                    created_entities['TenantEnrollment'] = updated_tenantenrollment
                else:
                    print('❌ Could not refresh tenantenrollment for update')
            except Exception as e:
                if 'version' in str(e).lower() or 'modified by another process' in str(e).lower():
                    print(
                        f'⚠️  tenantenrollment was modified by another process (optimistic locking): {e}'
                    )
                    print('💡 This is expected behavior in concurrent environments')
                else:
                    print(f'❌ Failed to update tenantenrollment: {e}')

        # 3. GET - Retrieve and print the entity
        if 'TenantEnrollment' in created_entities:
            print('\n🔍 Retrieving tenantenrollment...')
            try:
                entity_for_get = created_entities['TenantEnrollment']
                retrieved_tenantenrollment = self.tenantenrollment_repo.get_tenant_enrollment(
                    entity_for_get.tenant_id,
                    entity_for_get.user_id,
                    entity_for_get.course_id,
                    entity_for_get.enrollment_date,
                )

                if retrieved_tenantenrollment:
                    print(f'✅ Retrieved: {retrieved_tenantenrollment}')
                else:
                    print('❌ Failed to retrieve tenantenrollment')
            except Exception as e:
                print(f'❌ Failed to retrieve tenantenrollment: {e}')

        print('🎯 TenantEnrollment CRUD cycle completed!')

        # TenantLesson example
        print('\n--- TenantLesson ---')

        # 1. CREATE - Create sample tenantlesson
        sample_tenantlesson = TenantLesson(
            tenant_id='tenant-12345',
            course_id='course-11111',
            lesson_id='lesson-33333',
            lesson_order=1,
            title='Introduction to Variables',
            description='Learn about JavaScript variables and data types',
            content_type='video',
            content_url='https://example.com/videos/lesson-33333.mp4',
            duration_minutes=25,
            is_mandatory=True,
            quiz_required=True,
            passing_score=80,
            created_at=1704067200,
            updated_at=1705737600,
        )

        print('📝 Creating tenantlesson...')
        print(f'📝 PK: {sample_tenantlesson.pk()}, SK: {sample_tenantlesson.sk()}')

        try:
            created_tenantlesson = self.tenantlesson_repo.create_tenant_lesson(sample_tenantlesson)
            print(f'✅ Created: {created_tenantlesson}')
            # Store created entity for access pattern testing
            created_entities['TenantLesson'] = created_tenantlesson
        except Exception as e:
            # Check if the error is due to item already existing
            if 'ConditionalCheckFailedException' in str(e) or 'already exists' in str(e).lower():
                print('⚠️  tenantlesson already exists, retrieving existing entity...')
                try:
                    existing_tenantlesson = self.tenantlesson_repo.get_tenant_lesson(
                        sample_tenantlesson.tenant_id,
                        sample_tenantlesson.course_id,
                        sample_tenantlesson.lesson_order,
                        sample_tenantlesson.lesson_id,
                    )

                    if existing_tenantlesson:
                        print(f'✅ Retrieved existing: {existing_tenantlesson}')
                        # Store existing entity for access pattern testing
                        created_entities['TenantLesson'] = existing_tenantlesson
                    else:
                        print('❌ Failed to retrieve existing tenantlesson')
                except Exception as get_error:
                    print(f'❌ Failed to retrieve existing tenantlesson: {get_error}')
            else:
                print(f'❌ Failed to create tenantlesson: {e}')
        # 2. UPDATE - Update non-key field (title)
        if 'TenantLesson' in created_entities:
            print('\n🔄 Updating title field...')
            try:
                # Refresh entity to get latest version (handles optimistic locking)
                entity_for_refresh = created_entities['TenantLesson']
                refreshed_entity = self.tenantlesson_repo.get_tenant_lesson(
                    entity_for_refresh.tenant_id,
                    entity_for_refresh.course_id,
                    entity_for_refresh.lesson_order,
                    entity_for_refresh.lesson_id,
                )

                if refreshed_entity:
                    original_value = refreshed_entity.title
                    refreshed_entity.title = 'Introduction to Variables and Data Types'

                    updated_tenantlesson = self.tenantlesson_repo.update_tenant_lesson(
                        refreshed_entity
                    )
                    print(f'✅ Updated title: {original_value} → {updated_tenantlesson.title}')

                    # Update stored entity with updated values
                    created_entities['TenantLesson'] = updated_tenantlesson
                else:
                    print('❌ Could not refresh tenantlesson for update')
            except Exception as e:
                if 'version' in str(e).lower() or 'modified by another process' in str(e).lower():
                    print(
                        f'⚠️  tenantlesson was modified by another process (optimistic locking): {e}'
                    )
                    print('💡 This is expected behavior in concurrent environments')
                else:
                    print(f'❌ Failed to update tenantlesson: {e}')

        # 3. GET - Retrieve and print the entity
        if 'TenantLesson' in created_entities:
            print('\n🔍 Retrieving tenantlesson...')
            try:
                entity_for_get = created_entities['TenantLesson']
                retrieved_tenantlesson = self.tenantlesson_repo.get_tenant_lesson(
                    entity_for_get.tenant_id,
                    entity_for_get.course_id,
                    entity_for_get.lesson_order,
                    entity_for_get.lesson_id,
                )

                if retrieved_tenantlesson:
                    print(f'✅ Retrieved: {retrieved_tenantlesson}')
                else:
                    print('❌ Failed to retrieve tenantlesson')
            except Exception as e:
                print(f'❌ Failed to retrieve tenantlesson: {e}')

        print('🎯 TenantLesson CRUD cycle completed!')

        # TenantOrganization example
        print('\n--- TenantOrganization ---')

        # 1. CREATE - Create sample tenantorganization
        sample_tenantorganization = TenantOrganization(
            tenant_id='tenant-12345',
            organization_name='TechCorp Learning',
            domain='techcorp.com',
            subscription_plan='enterprise',
            max_users=500,
            max_courses=100,
            admin_email='admin@techcorp.com',
            created_at=1704067200,
            status='active',
        )

        print('📝 Creating tenantorganization...')
        print(f'📝 PK: {sample_tenantorganization.pk()}, SK: {sample_tenantorganization.sk()}')

        try:
            created_tenantorganization = self.tenantorganization_repo.create_tenant_organization(
                sample_tenantorganization
            )
            print(f'✅ Created: {created_tenantorganization}')
            # Store created entity for access pattern testing
            created_entities['TenantOrganization'] = created_tenantorganization
        except Exception as e:
            # Check if the error is due to item already existing
            if 'ConditionalCheckFailedException' in str(e) or 'already exists' in str(e).lower():
                print('⚠️  tenantorganization already exists, retrieving existing entity...')
                try:
                    existing_tenantorganization = (
                        self.tenantorganization_repo.get_tenant_organization(
                            sample_tenantorganization.tenant_id
                        )
                    )

                    if existing_tenantorganization:
                        print(f'✅ Retrieved existing: {existing_tenantorganization}')
                        # Store existing entity for access pattern testing
                        created_entities['TenantOrganization'] = existing_tenantorganization
                    else:
                        print('❌ Failed to retrieve existing tenantorganization')
                except Exception as get_error:
                    print(f'❌ Failed to retrieve existing tenantorganization: {get_error}')
            else:
                print(f'❌ Failed to create tenantorganization: {e}')
        # 2. UPDATE - Update non-key field (organization_name)
        if 'TenantOrganization' in created_entities:
            print('\n🔄 Updating organization_name field...')
            try:
                # Refresh entity to get latest version (handles optimistic locking)
                entity_for_refresh = created_entities['TenantOrganization']
                refreshed_entity = self.tenantorganization_repo.get_tenant_organization(
                    entity_for_refresh.tenant_id
                )

                if refreshed_entity:
                    original_value = refreshed_entity.organization_name
                    refreshed_entity.organization_name = 'TechCorp Learning Solutions'

                    updated_tenantorganization = (
                        self.tenantorganization_repo.update_tenant_organization(refreshed_entity)
                    )
                    print(
                        f'✅ Updated organization_name: {original_value} → {updated_tenantorganization.organization_name}'
                    )

                    # Update stored entity with updated values
                    created_entities['TenantOrganization'] = updated_tenantorganization
                else:
                    print('❌ Could not refresh tenantorganization for update')
            except Exception as e:
                if 'version' in str(e).lower() or 'modified by another process' in str(e).lower():
                    print(
                        f'⚠️  tenantorganization was modified by another process (optimistic locking): {e}'
                    )
                    print('💡 This is expected behavior in concurrent environments')
                else:
                    print(f'❌ Failed to update tenantorganization: {e}')

        # 3. GET - Retrieve and print the entity
        if 'TenantOrganization' in created_entities:
            print('\n🔍 Retrieving tenantorganization...')
            try:
                entity_for_get = created_entities['TenantOrganization']
                retrieved_tenantorganization = (
                    self.tenantorganization_repo.get_tenant_organization(entity_for_get.tenant_id)
                )

                if retrieved_tenantorganization:
                    print(f'✅ Retrieved: {retrieved_tenantorganization}')
                else:
                    print('❌ Failed to retrieve tenantorganization')
            except Exception as e:
                print(f'❌ Failed to retrieve tenantorganization: {e}')

        print('🎯 TenantOrganization CRUD cycle completed!')

        # TenantProgress example
        print('\n--- TenantProgress ---')

        # 1. CREATE - Create sample tenantprogress
        sample_tenantprogress = TenantProgress(
            tenant_id='tenant-12345',
            user_id='user-67890',
            course_id='course-11111',
            lesson_id='lesson-33333',
            attempt_date=1705737600,
            completion_status='completed',
            time_spent_minutes=30,
            quiz_score=92,
            quiz_passed=True,
            notes='Great progress on variables concept',
            last_accessed=1705824000,
        )

        print('📝 Creating tenantprogress...')
        print(f'📝 PK: {sample_tenantprogress.pk()}, SK: {sample_tenantprogress.sk()}')

        try:
            created_tenantprogress = self.tenantprogress_repo.create_tenant_progress(
                sample_tenantprogress
            )
            print(f'✅ Created: {created_tenantprogress}')
            # Store created entity for access pattern testing
            created_entities['TenantProgress'] = created_tenantprogress
        except Exception as e:
            # Check if the error is due to item already existing
            if 'ConditionalCheckFailedException' in str(e) or 'already exists' in str(e).lower():
                print('⚠️  tenantprogress already exists, retrieving existing entity...')
                try:
                    existing_tenantprogress = self.tenantprogress_repo.get_tenant_progress(
                        sample_tenantprogress.tenant_id,
                        sample_tenantprogress.user_id,
                        sample_tenantprogress.course_id,
                        sample_tenantprogress.lesson_id,
                        sample_tenantprogress.attempt_date,
                    )

                    if existing_tenantprogress:
                        print(f'✅ Retrieved existing: {existing_tenantprogress}')
                        # Store existing entity for access pattern testing
                        created_entities['TenantProgress'] = existing_tenantprogress
                    else:
                        print('❌ Failed to retrieve existing tenantprogress')
                except Exception as get_error:
                    print(f'❌ Failed to retrieve existing tenantprogress: {get_error}')
            else:
                print(f'❌ Failed to create tenantprogress: {e}')
        # 2. UPDATE - Update non-key field (completion_status)
        if 'TenantProgress' in created_entities:
            print('\n🔄 Updating completion_status field...')
            try:
                # Refresh entity to get latest version (handles optimistic locking)
                entity_for_refresh = created_entities['TenantProgress']
                refreshed_entity = self.tenantprogress_repo.get_tenant_progress(
                    entity_for_refresh.tenant_id,
                    entity_for_refresh.user_id,
                    entity_for_refresh.course_id,
                    entity_for_refresh.lesson_id,
                    entity_for_refresh.attempt_date,
                )

                if refreshed_entity:
                    original_value = refreshed_entity.completion_status
                    refreshed_entity.completion_status = 'mastered'

                    updated_tenantprogress = self.tenantprogress_repo.update_tenant_progress(
                        refreshed_entity
                    )
                    print(
                        f'✅ Updated completion_status: {original_value} → {updated_tenantprogress.completion_status}'
                    )

                    # Update stored entity with updated values
                    created_entities['TenantProgress'] = updated_tenantprogress
                else:
                    print('❌ Could not refresh tenantprogress for update')
            except Exception as e:
                if 'version' in str(e).lower() or 'modified by another process' in str(e).lower():
                    print(
                        f'⚠️  tenantprogress was modified by another process (optimistic locking): {e}'
                    )
                    print('💡 This is expected behavior in concurrent environments')
                else:
                    print(f'❌ Failed to update tenantprogress: {e}')

        # 3. GET - Retrieve and print the entity
        if 'TenantProgress' in created_entities:
            print('\n🔍 Retrieving tenantprogress...')
            try:
                entity_for_get = created_entities['TenantProgress']
                retrieved_tenantprogress = self.tenantprogress_repo.get_tenant_progress(
                    entity_for_get.tenant_id,
                    entity_for_get.user_id,
                    entity_for_get.course_id,
                    entity_for_get.lesson_id,
                    entity_for_get.attempt_date,
                )

                if retrieved_tenantprogress:
                    print(f'✅ Retrieved: {retrieved_tenantprogress}')
                else:
                    print('❌ Failed to retrieve tenantprogress')
            except Exception as e:
                print(f'❌ Failed to retrieve tenantprogress: {e}')

        print('🎯 TenantProgress CRUD cycle completed!')

        # TenantUser example
        print('\n--- TenantUser ---')

        # 1. CREATE - Create sample tenantuser
        sample_tenantuser = TenantUser(
            tenant_id='tenant-12345',
            user_id='user-67890',
            email='john.doe@techcorp.com',
            first_name='John',
            last_name='Doe',
            role='learner',
            department='Engineering',
            job_title='Software Developer',
            enrollment_date=1704931200,
            last_login=1705737600,
            status='active',
        )

        print('📝 Creating tenantuser...')
        print(f'📝 PK: {sample_tenantuser.pk()}, SK: {sample_tenantuser.sk()}')

        try:
            created_tenantuser = self.tenantuser_repo.create_tenant_user(sample_tenantuser)
            print(f'✅ Created: {created_tenantuser}')
            # Store created entity for access pattern testing
            created_entities['TenantUser'] = created_tenantuser
        except Exception as e:
            # Check if the error is due to item already existing
            if 'ConditionalCheckFailedException' in str(e) or 'already exists' in str(e).lower():
                print('⚠️  tenantuser already exists, retrieving existing entity...')
                try:
                    existing_tenantuser = self.tenantuser_repo.get_tenant_user(
                        sample_tenantuser.tenant_id, sample_tenantuser.user_id
                    )

                    if existing_tenantuser:
                        print(f'✅ Retrieved existing: {existing_tenantuser}')
                        # Store existing entity for access pattern testing
                        created_entities['TenantUser'] = existing_tenantuser
                    else:
                        print('❌ Failed to retrieve existing tenantuser')
                except Exception as get_error:
                    print(f'❌ Failed to retrieve existing tenantuser: {get_error}')
            else:
                print(f'❌ Failed to create tenantuser: {e}')
        # 2. UPDATE - Update non-key field (role)
        if 'TenantUser' in created_entities:
            print('\n🔄 Updating role field...')
            try:
                # Refresh entity to get latest version (handles optimistic locking)
                entity_for_refresh = created_entities['TenantUser']
                refreshed_entity = self.tenantuser_repo.get_tenant_user(
                    entity_for_refresh.tenant_id, entity_for_refresh.user_id
                )

                if refreshed_entity:
                    original_value = refreshed_entity.role
                    refreshed_entity.role = 'instructor'

                    updated_tenantuser = self.tenantuser_repo.update_tenant_user(refreshed_entity)
                    print(f'✅ Updated role: {original_value} → {updated_tenantuser.role}')

                    # Update stored entity with updated values
                    created_entities['TenantUser'] = updated_tenantuser
                else:
                    print('❌ Could not refresh tenantuser for update')
            except Exception as e:
                if 'version' in str(e).lower() or 'modified by another process' in str(e).lower():
                    print(
                        f'⚠️  tenantuser was modified by another process (optimistic locking): {e}'
                    )
                    print('💡 This is expected behavior in concurrent environments')
                else:
                    print(f'❌ Failed to update tenantuser: {e}')

        # 3. GET - Retrieve and print the entity
        if 'TenantUser' in created_entities:
            print('\n🔍 Retrieving tenantuser...')
            try:
                entity_for_get = created_entities['TenantUser']
                retrieved_tenantuser = self.tenantuser_repo.get_tenant_user(
                    entity_for_get.tenant_id, entity_for_get.user_id
                )

                if retrieved_tenantuser:
                    print(f'✅ Retrieved: {retrieved_tenantuser}')
                else:
                    print('❌ Failed to retrieve tenantuser')
            except Exception as e:
                print(f'❌ Failed to retrieve tenantuser: {e}')

        print('🎯 TenantUser CRUD cycle completed!')

        print('\n' + '=' * 50)
        print('🎉 Basic CRUD examples completed!')

        # Additional Access Pattern Testing Section (before cleanup)
        if include_additional_access_patterns:
            self._test_additional_access_patterns(created_entities)

        # Cleanup - Delete all created entities
        print('\n' + '=' * 50)
        print('🗑️  Cleanup: Deleting all created entities')
        print('=' * 50)

        # Delete TenantCertificate
        if 'TenantCertificate' in created_entities:
            print('\n🗑️  Deleting tenantcertificate...')
            try:
                deleted = self.tenantcertificate_repo.delete_tenant_certificate(
                    created_entities['TenantCertificate'].tenant_id,
                    created_entities['TenantCertificate'].user_id,
                    created_entities['TenantCertificate'].course_id,
                    created_entities['TenantCertificate'].issued_date,
                )

                if deleted:
                    print('✅ Deleted tenantcertificate successfully')
                else:
                    print('❌ Failed to delete tenantcertificate (not found or already deleted)')
            except Exception as e:
                print(f'❌ Failed to delete tenantcertificate: {e}')

        # Delete TenantCourse
        if 'TenantCourse' in created_entities:
            print('\n🗑️  Deleting tenantcourse...')
            try:
                deleted = self.tenantcourse_repo.delete_tenant_course(
                    created_entities['TenantCourse'].tenant_id,
                    created_entities['TenantCourse'].course_id,
                )

                if deleted:
                    print('✅ Deleted tenantcourse successfully')
                else:
                    print('❌ Failed to delete tenantcourse (not found or already deleted)')
            except Exception as e:
                print(f'❌ Failed to delete tenantcourse: {e}')

        # Delete TenantEnrollment
        if 'TenantEnrollment' in created_entities:
            print('\n🗑️  Deleting tenantenrollment...')
            try:
                deleted = self.tenantenrollment_repo.delete_tenant_enrollment(
                    created_entities['TenantEnrollment'].tenant_id,
                    created_entities['TenantEnrollment'].user_id,
                    created_entities['TenantEnrollment'].course_id,
                    created_entities['TenantEnrollment'].enrollment_date,
                )

                if deleted:
                    print('✅ Deleted tenantenrollment successfully')
                else:
                    print('❌ Failed to delete tenantenrollment (not found or already deleted)')
            except Exception as e:
                print(f'❌ Failed to delete tenantenrollment: {e}')

        # Delete TenantLesson
        if 'TenantLesson' in created_entities:
            print('\n🗑️  Deleting tenantlesson...')
            try:
                deleted = self.tenantlesson_repo.delete_tenant_lesson(
                    created_entities['TenantLesson'].tenant_id,
                    created_entities['TenantLesson'].course_id,
                    created_entities['TenantLesson'].lesson_order,
                    created_entities['TenantLesson'].lesson_id,
                )

                if deleted:
                    print('✅ Deleted tenantlesson successfully')
                else:
                    print('❌ Failed to delete tenantlesson (not found or already deleted)')
            except Exception as e:
                print(f'❌ Failed to delete tenantlesson: {e}')

        # Delete TenantOrganization
        if 'TenantOrganization' in created_entities:
            print('\n🗑️  Deleting tenantorganization...')
            try:
                deleted = self.tenantorganization_repo.delete_tenant_organization(
                    created_entities['TenantOrganization'].tenant_id
                )

                if deleted:
                    print('✅ Deleted tenantorganization successfully')
                else:
                    print('❌ Failed to delete tenantorganization (not found or already deleted)')
            except Exception as e:
                print(f'❌ Failed to delete tenantorganization: {e}')

        # Delete TenantProgress
        if 'TenantProgress' in created_entities:
            print('\n🗑️  Deleting tenantprogress...')
            try:
                deleted = self.tenantprogress_repo.delete_tenant_progress(
                    created_entities['TenantProgress'].tenant_id,
                    created_entities['TenantProgress'].user_id,
                    created_entities['TenantProgress'].course_id,
                    created_entities['TenantProgress'].lesson_id,
                    created_entities['TenantProgress'].attempt_date,
                )

                if deleted:
                    print('✅ Deleted tenantprogress successfully')
                else:
                    print('❌ Failed to delete tenantprogress (not found or already deleted)')
            except Exception as e:
                print(f'❌ Failed to delete tenantprogress: {e}')

        # Delete TenantUser
        if 'TenantUser' in created_entities:
            print('\n🗑️  Deleting tenantuser...')
            try:
                deleted = self.tenantuser_repo.delete_tenant_user(
                    created_entities['TenantUser'].tenant_id,
                    created_entities['TenantUser'].user_id,
                )

                if deleted:
                    print('✅ Deleted tenantuser successfully')
                else:
                    print('❌ Failed to delete tenantuser (not found or already deleted)')
            except Exception as e:
                print(f'❌ Failed to delete tenantuser: {e}')
        print('\n💡 Requirements:')
        print("   - DynamoDB table 'ELearningPlatform' must exist")
        print('   - DynamoDB permissions: GetItem, PutItem, UpdateItem, DeleteItem')

    def _test_additional_access_patterns(self, created_entities: dict):
        """Test additional access patterns beyond basic CRUD"""
        print('\n' + '=' * 60)
        print('🔍 Additional Access Pattern Testing')
        print('=' * 60)
        print()

        # TenantCertificate
        # Access Pattern #18: Get all certificates earned by user in tenant
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #18: Get all certificates earned by user in tenant')
            print('   Using Main Table')
            result = self.tenantcertificate_repo.get_user_certificates(
                created_entities['TenantCertificate'].tenant_id,
                created_entities['TenantCertificate'].user_id,
            )
            print('   ✅ Get all certificates earned by user in tenant completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #18: {e}')

        # Access Pattern #19: Issue certificate for course completion
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #19: Issue certificate for course completion')
            print('   Using Main Table')
            test_entity = TenantCertificate(
                tenant_id='sample_tenant_id',
                user_id='sample_user_id',
                course_id='sample_course_id',
                certificate_id='sample_certificate_id',
                course_title='sample_course_title',
                user_name='sample_user_name',
                instructor_name='sample_instructor_name',
                issued_date=0,
                completion_date=0,
                final_grade='sample_final_grade',
                certificate_url='sample_certificate_url',
                verification_code='sample_verification_code',
                expiry_date=0,
                status='sample_status',
            )
            result = self.tenantcertificate_repo.issue_course_certificate(test_entity)
            print('   ✅ Issue certificate for course completion completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #19: {e}')

        # Access Pattern #20: Verify certificate by verification code
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #20: Verify certificate by verification code')
            print('   Using Main Table')
            result = self.tenantcertificate_repo.verify_certificate(
                created_entities['TenantCertificate'].tenant_id,
                created_entities['TenantCertificate'].user_id,
                created_entities['TenantCertificate'].course_id,
                created_entities['TenantCertificate'].issued_date,
            )
            print('   ✅ Verify certificate by verification code completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #20: {e}')

        # TenantCourse
        # Access Pattern #6: Get course details within tenant
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #6: Get course details within tenant')
            print('   Using Main Table')
            result = self.tenantcourse_repo.get_tenant_course(
                created_entities['TenantCourse'].tenant_id,
                created_entities['TenantCourse'].course_id,
            )
            print('   ✅ Get course details within tenant completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #6: {e}')

        # Access Pattern #7: Create new course in tenant
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #7: Create new course in tenant')
            print('   Using Main Table')
            test_entity = TenantCourse(
                tenant_id='sample_tenant_id',
                course_id='sample_course_id',
                title='sample_title',
                description='sample_description',
                instructor_id='sample_instructor_id',
                instructor_name='sample_instructor_name',
                category='sample_category',
                difficulty_level='sample_difficulty_level',
                duration_hours=0,
                max_enrollments=0,
                prerequisites=['sample_prerequisite'],
                tags=['sample_tag'],
                created_at=0,
                updated_at=0,
                status='sample_status',
            )
            result = self.tenantcourse_repo.put_tenant_course(test_entity)
            print('   ✅ Create new course in tenant completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #7: {e}')

        # Access Pattern #8: Update course information
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #8: Update course information')
            print('   Using Main Table')
            result = self.tenantcourse_repo.update_course_details(
                created_entities['TenantCourse'].tenant_id,
                created_entities['TenantCourse'].course_id,
                created_entities['TenantCourse'],
            )
            print('   ✅ Update course information completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #8: {e}')

        # TenantEnrollment
        # Access Pattern #9: Get all course enrollments for a user in tenant
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #9: Get all course enrollments for a user in tenant')
            print('   Using Main Table')
            result = self.tenantenrollment_repo.get_user_enrollments(
                created_entities['TenantEnrollment'].tenant_id,
                created_entities['TenantEnrollment'].user_id,
            )
            print('   ✅ Get all course enrollments for a user in tenant completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #9: {e}')

        # Access Pattern #10: Enroll user in a course
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #10: Enroll user in a course')
            print('   Using Main Table')
            test_entity = TenantEnrollment(
                tenant_id='sample_tenant_id',
                user_id='sample_user_id',
                course_id='sample_course_id',
                course_title='sample_course_title',
                instructor_name='sample_instructor_name',
                enrollment_date=0,
                completion_date=42,
                progress_percentage=0,
                current_lesson='sample_current_lesson',
                grade='sample_grade',
                certificate_issued=False,
                status='sample_status',
            )
            result = self.tenantenrollment_repo.enroll_user_in_course(test_entity)
            print('   ✅ Enroll user in a course completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #10: {e}')

        # Access Pattern #11: Update user's progress in course
        # Index: Main Table
        try:
            print("🔍 Testing Access Pattern #11: Update user's progress in course")
            print('   Using Main Table')
            result = self.tenantenrollment_repo.update_enrollment_progress(
                created_entities['TenantEnrollment'].tenant_id,
                created_entities['TenantEnrollment'].user_id,
                created_entities['TenantEnrollment'].course_id,
                created_entities['TenantEnrollment'].enrollment_date,
                created_entities['TenantEnrollment'].progress_percentage,
                created_entities['TenantEnrollment'].current_lesson,
            )
            print("   ✅ Update user's progress in course completed")
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #11: {e}')

        # TenantLesson
        # Access Pattern #12: Get all lessons for a course in tenant (ordered)
        # Index: Main Table
        try:
            print(
                '🔍 Testing Access Pattern #12: Get all lessons for a course in tenant (ordered)'
            )
            print('   Using Main Table')
            result = self.tenantlesson_repo.get_course_lessons(
                created_entities['TenantLesson'].tenant_id,
                created_entities['TenantLesson'].course_id,
            )
            print('   ✅ Get all lessons for a course in tenant (ordered) completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #12: {e}')

        # Access Pattern #13: Get specific lesson details
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #13: Get specific lesson details')
            print('   Using Main Table')
            result = self.tenantlesson_repo.get_specific_lesson(
                created_entities['TenantLesson'].tenant_id,
                created_entities['TenantLesson'].course_id,
                created_entities['TenantLesson'].lesson_order,
                created_entities['TenantLesson'].lesson_id,
            )
            print('   ✅ Get specific lesson details completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #13: {e}')

        # Access Pattern #14: Create new lesson in course
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #14: Create new lesson in course')
            print('   Using Main Table')
            test_entity = TenantLesson(
                tenant_id='sample_tenant_id',
                course_id='sample_course_id',
                lesson_id='sample_lesson_id',
                lesson_order=0,
                title='sample_title',
                description='sample_description',
                content_type='sample_content_type',
                content_url='sample_content_url',
                duration_minutes=0,
                is_mandatory=False,
                quiz_required=False,
                passing_score=0,
                created_at=0,
                updated_at=0,
            )
            result = self.tenantlesson_repo.create_course_lesson(test_entity)
            print('   ✅ Create new lesson in course completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #14: {e}')

        # TenantOrganization
        # Access Pattern #1: Get organization details for a tenant
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #1: Get organization details for a tenant')
            print('   Using Main Table')
            result = self.tenantorganization_repo.get_tenant_organization(
                created_entities['TenantOrganization'].tenant_id
            )
            print('   ✅ Get organization details for a tenant completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #1: {e}')

        # Access Pattern #2: Create new tenant organization
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #2: Create new tenant organization')
            print('   Using Main Table')
            test_entity = TenantOrganization(
                tenant_id='sample_tenant_id',
                organization_name='sample_organization_name',
                domain='sample_domain',
                subscription_plan='sample_subscription_plan',
                max_users=0,
                max_courses=0,
                admin_email='sample_admin_email',
                created_at=0,
                status='sample_status',
            )
            result = self.tenantorganization_repo.put_tenant_organization(test_entity)
            print('   ✅ Create new tenant organization completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #2: {e}')

        # TenantProgress
        # Access Pattern #15: Get user's progress for all lessons in a course
        # Index: Main Table
        try:
            print("🔍 Testing Access Pattern #15: Get user's progress for all lessons in a course")
            print('   Using Main Table')
            result = self.tenantprogress_repo.get_user_course_progress(
                created_entities['TenantProgress'].tenant_id,
                created_entities['TenantProgress'].user_id,
                created_entities['TenantProgress'].course_id,
            )
            print("   ✅ Get user's progress for all lessons in a course completed")
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #15: {e}')

        # Access Pattern #16: Record user's progress on a lesson
        # Index: Main Table
        try:
            print("🔍 Testing Access Pattern #16: Record user's progress on a lesson")
            print('   Using Main Table')
            test_entity = TenantProgress(
                tenant_id='sample_tenant_id',
                user_id='sample_user_id',
                course_id='sample_course_id',
                lesson_id='sample_lesson_id',
                attempt_date=0,
                completion_status='sample_completion_status',
                time_spent_minutes=0,
                quiz_score=0,
                quiz_passed=False,
                notes='sample_notes',
                last_accessed=0,
            )
            result = self.tenantprogress_repo.record_lesson_progress(test_entity)
            print("   ✅ Record user's progress on a lesson completed")
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #16: {e}')

        # Access Pattern #17: Update user's lesson progress
        # Index: Main Table
        try:
            print("🔍 Testing Access Pattern #17: Update user's lesson progress")
            print('   Using Main Table')
            result = self.tenantprogress_repo.update_lesson_progress(
                created_entities['TenantProgress'].tenant_id,
                created_entities['TenantProgress'].user_id,
                created_entities['TenantProgress'].course_id,
                created_entities['TenantProgress'].lesson_id,
                created_entities['TenantProgress'].attempt_date,
                created_entities['TenantProgress'].completion_status,
                created_entities['TenantProgress'].time_spent_minutes,
            )
            print("   ✅ Update user's lesson progress completed")
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #17: {e}')

        # TenantUser
        # Access Pattern #3: Get user profile within tenant
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #3: Get user profile within tenant')
            print('   Using Main Table')
            result = self.tenantuser_repo.get_tenant_user(
                created_entities['TenantUser'].tenant_id, created_entities['TenantUser'].user_id
            )
            print('   ✅ Get user profile within tenant completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #3: {e}')

        # Access Pattern #4: Create new user in tenant
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #4: Create new user in tenant')
            print('   Using Main Table')
            test_entity = TenantUser(
                tenant_id='sample_tenant_id',
                user_id='sample_user_id',
                email='sample_email',
                first_name='sample_first_name',
                last_name='sample_last_name',
                role='sample_role',
                department='sample_department',
                job_title='sample_job_title',
                enrollment_date=0,
                last_login=0,
                status='sample_status',
            )
            result = self.tenantuser_repo.put_tenant_user(test_entity)
            print('   ✅ Create new user in tenant completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #4: {e}')

        # Access Pattern #5: Update user profile information
        # Index: Main Table
        try:
            print('🔍 Testing Access Pattern #5: Update user profile information')
            print('   Using Main Table')
            result = self.tenantuser_repo.update_user_profile(
                created_entities['TenantUser'].tenant_id,
                created_entities['TenantUser'].user_id,
                created_entities['TenantUser'],
            )
            print('   ✅ Update user profile information completed')
            print(f'   📊 Result: {result}')
        except Exception as e:
            print(f'❌ Error testing Access Pattern #5: {e}')

        print('\n💡 Access Pattern Implementation Notes:')
        print('   - Main Table queries use partition key and sort key')
        print('   - GSI queries use different key structures and may have range conditions')
        print(
            '   - Range conditions (begins_with, between, >, <, >=, <=) require additional parameters'
        )
        print('   - Implement the access pattern methods in your repository classes')


def main():
    """Main function to run examples"""
    # 🚨 SAFETY CHECK: Prevent accidental execution against production DynamoDB
    endpoint_url = os.getenv('AWS_ENDPOINT_URL_DYNAMODB', '')

    # Check if running against DynamoDB Local
    is_local = 'localhost' in endpoint_url.lower() or '127.0.0.1' in endpoint_url

    if not is_local:
        print('=' * 80)
        print('🚨 SAFETY WARNING: NOT RUNNING AGAINST DYNAMODB LOCAL')
        print('=' * 80)
        print()
        print(f'Current endpoint: {endpoint_url or "AWS DynamoDB (production)"}')
        print()
        print('⚠️  This script performs CREATE, UPDATE, and DELETE operations that could')
        print('   affect your production data!')
        print()
        print('To run against production DynamoDB:')
        print('  1. Review the code carefully to understand what data will be modified')
        print("  2. Search for 'SAFETY CHECK' in this file")
        print("  3. Comment out the 'raise RuntimeError' line below the safety check")
        print('  4. Understand the risks before proceeding')
        print()
        print('To run safely against DynamoDB Local:')
        print('  export AWS_ENDPOINT_URL_DYNAMODB=http://localhost:8000')
        print()
        print('=' * 80)

        # 🛑 SAFETY CHECK: Comment out this line to run against production
        raise RuntimeError(
            'Safety check: Refusing to run against production DynamoDB. See warning above.'
        )

    # Parse command line arguments
    include_additional_access_patterns = '--all' in sys.argv

    # Check if we're running against DynamoDB Local
    if endpoint_url:
        print(f'🔗 Using DynamoDB endpoint: {endpoint_url}')
        print(f'🌍 Using region: {os.getenv("AWS_DEFAULT_REGION", "us-east-1")}')
    else:
        print('🌐 Using AWS DynamoDB (no local endpoint specified)')

    print('📊 Using multiple tables:')
    print('   - ELearningPlatform')

    if include_additional_access_patterns:
        print('🔍 Including additional access pattern examples')

    examples = UsageExamples()
    examples.run_examples(include_additional_access_patterns=include_additional_access_patterns)


if __name__ == '__main__':
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Access pattern mapping and conflict resolution."""

from awslabs.dynamodb_mcp_server.repo_generation_tool.core.language_config import (
    LanguageConfig,
)
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.type_mappings import (
    TypeMapper,
)
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.utils import (
    filter_conflicting_patterns,
    get_crud_method_names,
    to_snake_case,
)
from typing import Any


class AccessPatternMapper:
    """Handles access pattern mapping and conflict resolution."""

    def __init__(self, language_config: LanguageConfig, type_mapper: TypeMapper = None):
        """Initialize the access pattern mapper."""
        self.language_config = language_config
        self.type_mapper = type_mapper

    def _get_equivalent_crud_method(
        self, pattern: dict, entity_name_snake: str, crud_methods: set[str]
    ) -> str:
        """Get the equivalent CRUD method name for a filtered pattern.

        Maps operations to their CRUD equivalents:
        - GetItem -> get_{entity}
        - PutItem -> create_{entity} (closest CRUD equivalent for conflict detection)
        - UpdateItem -> update_{entity}
        - DeleteItem -> delete_{entity}
        """
        operation = pattern.get('operation', '')

        # Map operation to CRUD method prefix
        operation_to_crud = {
            'GetItem': f'get_{entity_name_snake}',
            'PutItem': f'create_{entity_name_snake}',
            'UpdateItem': f'update_{entity_name_snake}',
            'DeleteItem': f'delete_{entity_name_snake}',
        }

        crud_method = operation_to_crud.get(operation)
        if crud_method and crud_method in crud_methods:
            return crud_method

        # Fallback to original name if no mapping found
        return pattern['name']

    def generate_mapping(
        self,
        entity_name: str,
        entity_config: dict[str, Any],
        gsi_list: list[dict[str, Any]] = None,
    ) -> dict[str, Any]:
        """Generate mapping for all access patterns in an entity.

        Args:
            entity_name: Name of the entity
            entity_config: Entity configuration dictionary
            gsi_list: Optional list of GSI definitions for projection info

        Returns:
            Dictionary mapping pattern IDs to pattern metadata including projection info
        """
        entity_name_snake = to_snake_case(entity_name)
        crud_methods = get_crud_method_names(entity_name, self.language_config)

        # Get the filtered/renamed patterns to determine actual method names
        filtered_patterns, _ = filter_conflicting_patterns(
            entity_config.get('access_patterns', []),
            crud_methods,
            entity_name=entity_name,
            entity_config=entity_config,
        )

        # Build a lookup from pattern_id to the resolved method name
        pattern_id_to_method = {}
        for pattern in filtered_patterns:
            pattern_id_to_method[pattern['pattern_id']] = pattern['name']

        # Generate mapping for all access patterns
        entity_mapping = {}
        for pattern in entity_config.get('access_patterns', []):
            pattern_id = str(pattern['pattern_id'])

            # Determine the actual method name (may be renamed or mapped to CRUD)
            if pattern['pattern_id'] in pattern_id_to_method:
                # Pattern was kept (possibly renamed)
                method_name = pattern_id_to_method[pattern['pattern_id']]
            else:
                # Pattern was filtered as duplicate - map to equivalent CRUD method
                method_name = self._get_equivalent_crud_method(
                    pattern, entity_name_snake, crud_methods
                )

            # Get the actual return type based on operation and return_type
            schema_return_type = pattern.get('return_type', 'Any')
            operation = pattern.get('operation', 'Unknown')

            # For Query/Scan operations returning entity_list, use paginated return type
            if (
                self.type_mapper
                and operation in ['Query', 'Scan']
                and schema_return_type == 'entity_list'
            ):
                actual_return_type = f'tuple[list[{entity_name}], dict | None]'
            # For Query/Scan operations returning mixed_data (item collections), use paginated dict return type
            elif (
                self.type_mapper
                and operation in ['Query', 'Scan']
                and schema_return_type == 'mixed_data'
            ):
                actual_return_type = 'tuple[list[dict[str, Any]], dict | None]'
            elif self.type_mapper:
                actual_return_type = self.type_mapper.map_return_type(
                    schema_return_type, entity_name
                )
            else:
                actual_return_type = schema_return_type

            mapping_entry = {
                'pattern_id': pattern['pattern_id'],
                'description': pattern['description'],
                'entity': entity_name,
                'repository': f'{entity_name}Repository',
                'method_name': method_name,
                'parameters': pattern.get('parameters', []),
                'return_type': actual_return_type,
                'operation': operation,
                'index_name': pattern.get('index_name'),
                'range_condition': pattern.get('range_condition'),
            }

            # Include consistent_read for read operations (GetItem, Query, Scan, BatchGetItem)
            # Defaults to false (eventually consistent). Omit for write operations.
            read_operations = {'GetItem', 'Query', 'Scan', 'BatchGetItem'}
            if operation in read_operations:
                mapping_entry['consistent_read'] = pattern.get('consistent_read', False)

            entity_mapping[pattern_id] = mapping_entry

            # Add GSI projection info if this pattern uses a GSI
            if pattern.get('index_name') and gsi_list:
                # Find the GSI definition
                gsi = next((g for g in gsi_list if g.get('name') == pattern['index_name']), None)
                if gsi:
                    entity_mapping[pattern_id]['projection'] = gsi.get('projection', 'ALL')
                    if gsi.get('projection') == 'INCLUDE' and 'included_attributes' in gsi:
                        entity_mapping[pattern_id]['projected_attributes'] = gsi[
                            'included_attributes'
                        ]

        return entity_mapping



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Loader for structured usage data from JSON files."""

import logging
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.file_utils import (
    FileUtils,
)
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.usage_data_formatter import (
    UsageDataFormatterInterface,
)
from pathlib import Path
from typing import Any, Dict, Optional


logger = logging.getLogger(__name__)


class UsageDataLoader:
    """Loads realistic sample data from structured JSON files."""

    def __init__(
        self,
        usage_data_path: Optional[str] = None,
        formatter: Optional[UsageDataFormatterInterface] = None,
    ):
        """Initialize loader with optional usage data file path and formatter.

        Args:
            usage_data_path: Path to usage data JSON file
            formatter: Language-specific formatter for values
        """
        self.usage_data_path = usage_data_path
        self.usage_data = {}
        self.formatter = formatter

        if usage_data_path:
            usage_path = Path(usage_data_path)
            if usage_path.exists():
                self._load_usage_data()

    def get_sample_value_for_field(
        self,
        field_name: str,
        field_type: str,
        entity_name: Optional[str] = None,
        use_access_pattern_data: bool = False,
    ) -> Optional[str]:
        """Get a realistic sample value for a field based on the usage data.

        Lookup hierarchy:
        1. Entity-specific sample_data or access_pattern_data
        2. Return None if not found
        """
        if not self.formatter:
            return None

        # Try entity-specific data
        if entity_name:
            entities = self.usage_data.get('entities', {})
            if entity_name in entities:
                # Use access_pattern_data if requested, otherwise sample_data
                data_key = 'access_pattern_data' if use_access_pattern_data else 'sample_data'
                sample_data = entities[entity_name].get(data_key, {})
                if field_name in sample_data:
                    return self.formatter.format_value(sample_data[field_name], field_type)

        return None

    def get_update_value_for_field(
        self, field_name: str, field_type: str, entity_name: Optional[str] = None
    ) -> Optional[str]:
        """Get a realistic update value for a field.

        Lookup hierarchy:
        1. Entity-specific update_data
        2. Return None if not found
        """
        if not self.formatter:
            return None

        # Try entity-specific update data
        if entity_name:
            entities = self.usage_data.get('entities', {})
            if entity_name in entities:
                update_data = entities[entity_name].get('update_data', {})
                if field_name in update_data:
                    return self.formatter.format_value(update_data[field_name], field_type)

        return None

    def get_all_usage_data(self) -> Dict[str, Any]:
        """Get all loaded usage data."""
        return self.usage_data

    def has_data(self) -> bool:
        """Check if any usage data was successfully loaded."""
        return bool(self.usage_data)

    def get_entity_sample_data(self, entity_name: str) -> Dict[str, Any]:
        """Get all sample data for a specific entity."""
        entities = self.usage_data.get('entities', {})
        return entities.get(entity_name, {}).get('sample_data', {})

    def get_entity_update_data(self, entity_name: str) -> Dict[str, Any]:
        """Get all update data for a specific entity."""
        entities = self.usage_data.get('entities', {})
        return entities.get(entity_name, {}).get('update_data', {})

    def _load_usage_data(self) -> None:
        """Load the usage data JSON file."""
        try:
            self.usage_data = FileUtils.load_json_file(self.usage_data_path, 'Usage data')
            logger.info(f'Successfully loaded usage data from {self.usage_data_path}')
        except (FileNotFoundError, ValueError) as e:
            logger.warning(f'Could not load usage data file {self.usage_data_path}: {e}')
            self.usage_data = {}
        except Exception as e:
            logger.warning(f'Unexpected error loading usage data file {self.usage_data_path}: {e}')
            self.usage_data = {}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base generator abstract class and core interfaces."""

from abc import ABC, abstractmethod
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.language_config import (
    LanguageConfigLoader,
)
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.schema_loader import SchemaLoader
from awslabs.dynamodb_mcp_server.repo_generation_tool.core.type_mappings import TypeMapper
from typing import Any


class BaseGenerator(ABC):
    """Base class for code generators."""

    def __init__(self, schema_path: str, language: str = 'python'):
        """Initialize code generator.

        Args:
            schema_path: Path to schema file
            language: Target programming language
        """
        self.schema_loader = SchemaLoader(schema_path)
        self.language = language
        self.language_config = LanguageConfigLoader.load(language)
        self.type_mapper = TypeMapper(language)

    @property
    def schema(self) -> dict[str, Any]:
        """Get the loaded schema."""
        return self.schema_loader.schema

    @abstractmethod
    def generate_entity(self, entity_name: str, entity_config: dict[str, Any]) -> str:
        """Generate entity code."""
        pass

    @abstractmethod
    def generate_repository(self, entity_name: str, entity_config: dict[str, Any]) -> str:
        """Generate repository code."""
        pass

    @abstractmethod
    def generate_all(self, output_dir: str, generate_usage_examples: bool = False) -> None:
        """Generate all code artifacts."""
        pass

    def generate_usage_examples(
        self,
        access_pattern_mapping: dict[str, Any],
        all_entities: dict[str, Any],
        all_tables: list[dict[str, Any]],
    ) -> str:
        """Generate usage examples code - to be implemented by subclasses."""
        return '# Usage examples not implemented for this generator'
